/*
Krake Webcrawlerframework
Copyright (C) LS2 TU Dortmund (http://ls2-www.cs.tu-dortmund.de)

This file is part of the Krake Webcrawlerframework.
The sourcecode and data files provided come without any warranty. Use at your own risk!

The sourcecode is the intellectual property of its authors. You have permission to download
and use this sourcecode for the sake of inspection and for private and academic use. You are
not allowed to redistribute this sourcecode without our permission. If you want to use the
sourcecode in any other way as granted above,
please contact Christian Sohler <christian.sohler[at]tu-dortmund.de>.
*/
package krake;

import org.apache.hadoop.fs.FileStatus;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapreduce.Job;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.security.MessageDigest;
import java.util.HashSet;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Collection of commonly used methods and constants.
 * @author Thomas Nitschke, Alex Munteanu, Christopher Schröder, Marc Gille
 *
 */
public class Common
{

	// helper for doExport
	static void collectInputs(List<Path> paths, FileSystem fs, Path cur) throws IOException
	{
		FileStatus[] entries = fs.listStatus(cur);
		Pattern partpt = Pattern.compile("part-m-\\d\\d\\d\\d\\d");
		
		for(FileStatus entry : entries)
		{
			// recurse into subdirectories
			if(entry.isDir())
				collectInputs(paths, fs, entry.getPath());
			else // add normal files
			{
				// ignore temporary hadoop outputs/logs
				Matcher m = partpt.matcher(entry.getPath().getName());
				if(m.matches())
					paths.add(entry.getPath());
			}
		}
	}

	public static org.apache.hadoop.conf.Configuration getDefaultHadoopConfig() throws IOException
	{
		Job j = new Job();
		return j.getConfiguration();
	}

	/**
	 * Ids for various counters that the jobs use.
	 */
	public enum CrawlCounters
	{
		SEGMENTS, MAPPERS, BLOCKED_BY_ROBOTS, BLOCKED_BY_FILTER, BLOCKED_BY_ERROR, BLOCKED_BY_HEADER, BLOCKED_BY_TYPE, BLOCKED_BY_SIZE, BLOCKED_BY_TIMEOUT, BLOCKED_BY_NON_LI, WRITTEN, NEW_URLS, SEEN_URLS, INPUT_URLS, ALREADY_CRAWLED, DELETE_ROW, DELETE_LINK, BLOCKED_BY_REDIRECT, BLOCKED_BY_INVALID_URL, UNMAPPED_URLS, PAGES, UNPARSABLE
	}

	// / pattern for href-links
	public static Pattern LinkPattern = Pattern.compile("<\\s*a.*\\s+href\\s*=\\s*(?:'(.+?)'|\"(.+?)\")");

	// / pattern for frames and iframe src-links
	public static Pattern FramePattern = Pattern.compile("<\\s*(?:frame|iframe).*\\s+src\\s*=\\s*(?:'(.+?)'|\"(.+?)\")");

	/**
	 * Converts a byte-Array to a hexstring
	 * 
	 * @param a
	 *            Array of bytes
	 * @return Hexstring without 0x-prefix
	 */
	public static String byteArrayToHexString(byte[] a)
	{
		StringBuilder b = new StringBuilder();
		final char[] table = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F' };
		int x;

		for(int i = 0; i < a.length; i++)
		{
			x = a[i] & 0xFF;

			char f = table[x >> 4];
			char s = table[x & 0x0F];
			b.append(f);
			b.append(s);
		}
		return b.toString();
	}

	public static Text byteArrayToHex(byte[] a)
	{
		return new Text(byteArrayToHexString(a));
	}

	/**
	 * Converts a hexstring to a byte-Array
	 * 
	 * @param t
	 *            Hexstring (without 0x-prefix)
	 * @return Bytes represented by t
	 */
	public static byte[] HextoByteArray(String t)
	{
		String s = t.toUpperCase();

		byte[] buff = new byte[(s.length() % 2) == 0 ? s.length() / 2 : s.length() / 2 + 1];
		int wi = buff.length - 1;
		int shift;

		for(int i = s.length() - 1; i >= 0; --i)
		{
			if(((s.length() - i + 1) % 2) == 0)
			{
				buff[wi] = 0;
				shift = 0;
			}
			else
				shift = 4;

			switch(s.charAt(i))
			{
			case '0':
				break;
			case '1':
				buff[wi] |= (0x1 << shift);
				break;
			case '2':
				buff[wi] |= (0x2 << shift);
				break;
			case '3':
				buff[wi] |= (0x3 << shift);
				break;
			case '4':
				buff[wi] |= (0x4 << shift);
				break;
			case '5':
				buff[wi] |= (0x5 << shift);
				break;
			case '6':
				buff[wi] |= (0x6 << shift);
				break;
			case '7':
				buff[wi] |= (0x7 << shift);
				break;
			case '8':
				buff[wi] |= (0x8 << shift);
				break;
			case '9':
				buff[wi] |= (0x9 << shift);
				break;
			case 'A':
				buff[wi] |= (0xA << shift);
				break;
			case 'B':
				buff[wi] |= (0xB << shift);
				break;
			case 'C':
				buff[wi] |= (0xC << shift);
				break;
			case 'D':
				buff[wi] |= (0xD << shift);
				break;
			case 'E':
				buff[wi] |= (0xE << shift);
				break;
			case 'F':
				buff[wi] |= (0xF << shift);
				break;
			default:
				throw new IllegalArgumentException("Not a hex string: '" + s + "' with char '" + s.charAt(i) + "' at " + i);
			}

			if(((s.length() - i + 1) % 2) == 1)
				--wi;
		}
		return buff;
	}

	/**
	 * Checks if the file-extension of the URL indicates a filetype that does not contain any textdata.
	 * 
	 * @param u
	 *            URL
	 * @return true, if the URL does not contain any textdata.
	 */
	public static boolean isNonTextURL(URL u)
	{
		// also filters rss feeds and related xml stuff
		final String[] endings = { ".swf", ".atom", ".rss", ".rss2", ".rdf", ".txt", ".xml", ".7z", ".ace", ".avi", ".bmp", ".bz2", ".cab",
				".css", ".djvu", ".dll", ".doc", ".docx", ".exe", ".flv", ".gif", ".gz", ".ico", ".img", ".iso", ".iso9", ".jar", ".java",
				".jpeg", ".jpg", ".js", ".m1v", ".m4v", ".md5", ".mov", ".mp3", ".mp4", ".mpa", ".mpe", ".mpeg", ".mpg", ".msi", ".ogg",
				".pdf", ".png", ".pps", ".ppt", ".ps", ".rar", ".rss", ".rss2", ".rtf", ".sha1", ".sha512", ".tar", ".tex", ".tgz", ".tif",
				".txt", ".wav", ".wma", ".wml", ".wmv", ".xls", ".xml", ".xpm", ".xs", ".xsd", ".xsl", ".zip", ".psd", ".tiff" };
		String file = u.getFile().toLowerCase();
		for(String e : endings)
		{
			if(file.endsWith(e))
			{
				return true;
			}
		}
		return false;
	}

	/**
	 * Normalize URL on the basis of
	 * http://en.wikipedia.org/wiki/URL_normalization and return sha1 hash of the normalized URL.
	 * 
	 * @return SHA1 Hash of the normalized URL
	 */
	public static String normalize(URL url)
	{
		URL cleanurl;
		try
		{
			cleanurl = new URL(killJSession(url.toString()));
		}
		catch(MalformedURLException e)
		{
			e.printStackTrace();
			return null;
		}
		// convert protocol and host to lowercase
		String protocol = cleanurl.getProtocol().toLowerCase();
		String host = cleanurl.getHost().toLowerCase();
		String path = cleanurl.getPath();
		// remove "www." prefix
		if(host.startsWith("www."))
		{
			host = host.substring(4);
		}
		// remove relative paths
		if(path.contains("../"))
		{
			// System.out.println(path);
			path = path.replaceAll("([^/]*)/\\.\\./", "");
		}
		if(path.contains("./"))
		{
			path = path.replaceAll("\\./", "");
		}

		// remove default directory index
		if(path.toLowerCase().endsWith("index.htm"))
		{
			path = path.substring(0, path.length() - 9);
		}
		if(path.toLowerCase().endsWith("index.html"))
		{
			path = path.substring(0, path.length() - 10);
		}
		if(path.toLowerCase().endsWith("index.shtm"))
		{
			path = path.substring(0, path.length() - 10);
		}
		if(path.toLowerCase().endsWith("index.shtml"))
		{
			path = path.substring(0, path.length() - 11);
		}
		if(path.toLowerCase().endsWith("index.xhtm"))
		{
			path = path.substring(0, path.length() - 10);
		}
		if(path.toLowerCase().endsWith("index.xhtml"))
		{
			path = path.substring(0, path.length() - 11);
		}
		if(path.toLowerCase().endsWith("index.php"))
		{
			path = path.substring(0, path.length() - 9);
		}
		if(path.toLowerCase().endsWith("index.cgi"))
		{
			path = path.substring(0, path.length() - 9);
		}
		if(path.toLowerCase().endsWith("index.cfm"))
		{
			path = path.substring(0, path.length() - 9);
		}
		if(path.toLowerCase().endsWith("index.pl"))
		{
			path = path.substring(0, path.length() - 8);
		}
		if(path.toLowerCase().endsWith("index.jsp"))
		{
			path = path.substring(0, path.length() - 9);
		}
		if(path.toLowerCase().endsWith("index.asp"))
		{
			path = path.substring(0, path.length() - 9);
		}
		if(path.toLowerCase().endsWith("index.aspx"))
		{
			path = path.substring(0, path.length() - 10);
		}
		if(path.toLowerCase().endsWith("index.php4"))
		{
			path = path.substring(0, path.length() - 10);
		}
		if(path.toLowerCase().endsWith("index.php5"))
		{
			path = path.substring(0, path.length() - 10);
		}
		if(path.toLowerCase().endsWith("index.lasso"))
		{
			path = path.substring(0, path.length() - 11);
		}
		if(path.toLowerCase().endsWith("default.php"))
		{
			path = path.substring(0, path.length() - 11);
		}
		if(path.toLowerCase().endsWith("default.htm"))
		{
			path = path.substring(0, path.length() - 11);
		}
		if(path.toLowerCase().endsWith("default.html"))
		{
			path = path.substring(0, path.length() - 12);
		}
		if(path.toLowerCase().endsWith("default.asp"))
		{
			path = path.substring(0, path.length() - 11);
		}
		if(path.toLowerCase().endsWith("default.aspx"))
		{
			path = path.substring(0, path.length() - 12);
		}

		String[] items = path.split("/");
		if(items.length > 0)
		{
			String last_item = items[0];
			int cnt = 1;
			path = items[0];
			for(int i = 1; i < items.length; ++i)
			{
				if(items[i].equals(last_item))
					++cnt;
				else
				{
					cnt = 1;
					last_item = items[i];
				}
				if(cnt <= 2)
					path += "/" + items[i];
			}
		}

		// add trailing slash
		if(!path.startsWith("/") || (!path.endsWith("/") && (path.lastIndexOf("/") > path.lastIndexOf("."))))
		{
			path = path + "/";
		}

		// sort query string
		String query = cleanurl.getQuery();
		if(query != null)
		{
			query = query.replaceAll("amp%3B", "&");
			query = query.replaceAll("amp%3b", "&");
			query = query.replaceAll("amp;", "&");
			query = query.replaceAll("%26", "&");
			query = query.replaceAll("%3D", "=");
			query = query.replaceAll("%3d", "=");

			String[] splitQuery = query.split("&");
			if(splitQuery.length > 0)
			{
				HashSet<String> toks = new HashSet<String>();
				for(String item : splitQuery)
					toks.add(item);

				splitQuery = toks.toArray(new String[0]);

				java.util.Arrays.sort(splitQuery);
				query = "";
				for(int i = 0; i < splitQuery.length; i++)
				{
					if(splitQuery[i].toLowerCase().startsWith("sid="))
					{
						continue;
					}
					if(splitQuery[i].toLowerCase().startsWith("sessid="))
					{
						continue;
					}
					if(splitQuery[i].toLowerCase().startsWith("phpsessid="))
					{
						continue;
					}
					if(splitQuery[i].toLowerCase().startsWith("phpkitsid="))
					{
						continue;
					}
					if(splitQuery[i].toLowerCase().startsWith("jsessionid="))
					{
						continue;
					}
					if(splitQuery[i].toLowerCase().startsWith("psid="))
					{
						continue;
					}
					if(splitQuery[i].toLowerCase().startsWith("session_id="))
					{
						continue;
					}
					if(splitQuery[i].toLowerCase().startsWith("msid="))
					{
						continue;
					}
					if(splitQuery[i].toLowerCase().startsWith("oscsid="))
					{
						continue;
					}
					if(splitQuery[i].toLowerCase().startsWith("bigwarecsid="))
					{
						continue;
					}
					if(splitQuery[i].toLowerCase().startsWith("xtcsid="))
					{
						continue;
					}
					if(splitQuery[i].toLowerCase().startsWith("wuid="))
					{
						continue;
					}
					if(splitQuery[i].toLowerCase().startsWith("g2_gallerysid="))
					{
						continue;
					}
					if(splitQuery[i].toLowerCase().startsWith("basket_id="))
					{
						continue;
					}
					if(splitQuery[i].toLowerCase().startsWith("randomstamp="))
					{
						continue;
					}
					if(splitQuery[i].toLowerCase().startsWith("wkid="))
					{
						continue;
					}
					if(splitQuery[i].toLowerCase().contains("xtcsid="))
					{
						splitQuery[i] = splitQuery[i].substring(0, splitQuery[i].toLowerCase().indexOf("xtcsid="));
					}
					if(splitQuery[i].toLowerCase().startsWith("el_mcal_year="))
					{
						try
						{
							if(Integer.parseInt(splitQuery[i].substring(13)) > 2020)
								continue;
						}
						catch(Exception e)
						{
						}
					}
					if(splitQuery[i].toLowerCase().equals(""))
					{
						continue;
					}
					if(query.equals(""))
						query = splitQuery[i];
					else
						query += "&" + splitQuery[i];
				}
			}
		}

		String file = (query == null || query.compareTo("") == 0) ? path : path + "?" + query;

		// capitalize letters in escape sequences
		String finalFile = "";
		for(int i = 0; i < file.length(); i++)
		{
			if(file.charAt(i) == '%' && i + 2 < file.length() && isHEX(file.substring(i + 1, i + 3)))
			{
				finalFile += Character.toUpperCase(file.charAt(i + 1));
				finalFile += Character.toUpperCase(file.charAt(i + 2));
				i += 2; // jump over the two hex chars
			}
			else
			{
				finalFile += file.charAt(i);
			}
		}
		file = finalFile;

		URL ret = null;
		try
		{
			// removing default port
			if(cleanurl.getPort() == -1 || cleanurl.getPort() == 80)
			{
				ret = new URL(protocol, host, file);
			}
			else
			{
				ret = new URL(protocol, host, cleanurl.getPort(), file);
			}
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}

		// System.out.println(ret.toString());

		byte[] hash = null;
		try
		{
			MessageDigest sha1 = MessageDigest.getInstance("SHA1");
			hash = sha1.digest(ret.toString().getBytes());
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}

		return byteArrayToHexString(hash);
	}

	/**
	 * @param hex
	 *            hexstring
	 * @return true if the supplied string is a hexadecimal representation of a
	 *         32 bit integer.
	 */
	public static boolean isHEX(String hex)
	{
		try
		{
			Integer.parseInt(hex, 16);
		}
		catch(Exception e)
		{
			return false;
		}
		return true;
	}

	/**
	 * @param name
	 *            URL-String
	 * @return true if the URL-Protocoll is supported by the crawler
	 */
	public static boolean isSupportedProtocol(String name)
	{
		return name.equals("http");// https is not supported due to cretificate issues
	}

	/**
	 * Converts an array of 4 bytes to a signed integer.
	 * @param b byte array of length 4
	 * @return Integer represented by b
	 */
	public static final int intFromBytes(byte[] b)
	{
		int l = 0;
		l |= b[0] & 0xFF;
		l <<= 8;
		l |= b[1] & 0xFF;
		l <<= 8;
		l |= b[2] & 0xFF;
		l <<= 8;
		l |= b[3] & 0xFF;
		return l;
	}

	/**
	 * Removes jsessionIDs froman url.
	 * @param url string to be cleaned
	 * @return clean url
	 */
	private static String killJSession(String url)
	{
		Matcher m = pt_jsession1.matcher(url);
		m = pt_jsession2.matcher(m.replaceAll("?"));
		m = pt_jsession3.matcher(m.replaceAll("%3f"));
		return m.replaceAll("");
	}

	private static Pattern pt_jsession1 = Pattern.compile(";jsessionid=[a-zA-Z0-9]+\\??");
	private static Pattern pt_jsession2 = Pattern.compile("%3(b|B)jsessionid%3(d|D)[a-zA-Z0-9]+(%3(f|F))?");
	private static Pattern pt_jsession3 = Pattern.compile("(\\?|%3f)$");
}
