/*
Krake Webcrawlerframework
Copyright (C) LS2 TU Dortmund (http://ls2-www.cs.tu-dortmund.de)

This file is part of the Krake Webcrawlerframework.
The sourcecode and data files provided come without any warranty. Use at your own risk!

The sourcecode is the intellectual property of its authors. You have permission to download
and use this sourcecode for the sake of inspection and for private and academic use. You are
not allowed to redistribute this sourcecode without our permission. If you want to use the
sourcecode in any other way as granted above,
please contact Christian Sohler <christian.sohler[at]tu-dortmund.de>.
*/
package krake;

import krake.mapreduce.*;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.conf.Configured;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapreduce.lib.input.SequenceFileInputFormat;
import org.apache.hadoop.mapreduce.lib.output.NullOutputFormat;
import org.apache.hadoop.util.Tool;
import org.apache.hadoop.util.ToolRunner;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

/**
 * Central job-management class for the krake-crawler. All jobs, including
 * non-hadoop jobs are started through this class.
 *
 * @author Thomas Nitschke
 */
public class Exporter extends Configured implements Tool
{
	private final Log LOG = LogFactory.getLog(Exporter.class);

	@Override
	public int run(String[] args) throws Exception
	{
		KrakeConfig.load();

		int pagecount = runCounter();
		LOG.info("Counted " + pagecount + " pages for export, starting to export...");
		runExporter(pagecount);
		return 0;
	}

	private int runCounter() throws ClassNotFoundException, IOException, InterruptedException
	{
		getConf().set("mapred.child.java.opts", "-Xmx2000m"); // allow crawlmappers to use more memory
		org.apache.hadoop.mapreduce.Job job = new org.apache.hadoop.mapreduce.Job(getConf(), "CounterJob");
		job.setJarByClass(Exporter.class);
		job.setMapperClass(ExportCounterMapper.class);


		// setup input
		job.setInputFormatClass(SequenceFileInputFormat.class);

		// collect all input files (everything in the crawl base directory)
		FileSystem fs = job.getWorkingDirectory().getFileSystem(job.getConfiguration());
		List<Path> paths = new ArrayList<Path>(32);
		Common.collectInputs(paths, fs, new Path(KrakeConfig.crawl_base));

		for(Path p : paths)
			SequenceFileInputFormat.addInputPath(job, p);

		// setup output
		Configuration conf = job.getConfiguration();
		conf.set(ExportReducer.PATH_CONTENT, KrakeConfig.content_path);
		conf.set(ExportReducer.PATH_OFFSETS, KrakeConfig.offsets_path);
		conf.set(ExportReducer.PATH_HASH_TO_URL, KrakeConfig.hash_to_url_path);
		conf.set(ExportReducer.PATH_ID_TO_HASH, KrakeConfig.id_to_hash_path);

		// reducer does manual output
		job.setOutputFormatClass(NullOutputFormat.class);

		// only a single reducer allowed
		job.setReducerClass(ExportCounterReducer.class);
		job.setMapOutputKeyClass(Text.class);
		job.setMapOutputValueClass(IntWritable.class);

		job.submit();
		job.waitForCompletion(true);

		if(!job.isSuccessful())
			throw new IOException("Job failed");


		int pagecount = (int)job.getCounters().findCounter(Common.CrawlCounters.PAGES).getValue();
		return pagecount;
	}

	private void runExporter(int pagecount) throws IOException, ClassNotFoundException, InterruptedException
	{
		getConf().set("mapred.child.java.opts", "-Xmx2000m"); // allow crawlmappers to use more memory
		org.apache.hadoop.mapreduce.Job job = new org.apache.hadoop.mapreduce.Job(getConf(), "ExportJob");
		job.setJarByClass(Exporter.class);
		job.setMapperClass(ExportMapper.class);

		// setup input
		job.setInputFormatClass(SequenceFileInputFormat.class);

		// collect all input files (everything in the crawl base directory)
		FileSystem fs = job.getWorkingDirectory().getFileSystem(job.getConfiguration());
		List<Path> paths = new ArrayList<Path>(32);
		Common.collectInputs(paths, fs, new Path(KrakeConfig.crawl_base));

		for(Path p : paths)
			SequenceFileInputFormat.addInputPath(job, p);

		// setup output
		Configuration conf = job.getConfiguration();
		conf.set(ExportReducer.PATH_CONTENT, KrakeConfig.content_path);
		conf.set(ExportReducer.PATH_OFFSETS, KrakeConfig.offsets_path);
		conf.set(ExportReducer.PATH_HASH_TO_URL, KrakeConfig.hash_to_url_path);
		conf.set(ExportReducer.PATH_ID_TO_HASH, KrakeConfig.id_to_hash_path);
		conf.setInt(ExportReducer.PAGE_COUNT, pagecount);

		// reducer does manual output
		job.setOutputFormatClass(NullOutputFormat.class);

		// only a single reducer allowed
		job.setNumReduceTasks(1);
		job.setReducerClass(ExportReducer.class);
		job.setMapOutputKeyClass(Text.class);
		job.setMapOutputValueClass(TextPairWritable.class);

		job.submit();
		job.waitForCompletion(true);
		if(!job.isSuccessful())
			throw new IOException("Job failed");
	}

	public static void main(String[] args) throws Exception
	{
		int ret = ToolRunner.run(new Exporter(), args);
		System.exit(ret);
	}
}

