/*
Krake Webcrawlerframework
Copyright (C) LS2 TU Dortmund (http://ls2-www.cs.tu-dortmund.de)

This file is part of the Krake Webcrawlerframework.
The sourcecode and data files provided come without any warranty. Use at your own risk!

The sourcecode is the intellectual property of its authors. You have permission to download
and use this sourcecode for the sake of inspection and for private and academic use. You are
not allowed to redistribute this sourcecode without our permission. If you want to use the
sourcecode in any other way as granted above,
please contact Christian Sohler <christian.sohler[at]tu-dortmund.de>.
*/
package krake;

import krake.file.URLMapping;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.conf.Configured;
import org.apache.hadoop.fs.FSDataInputStream;
import org.apache.hadoop.fs.FSDataOutputStream;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.mapreduce.Job;
import org.apache.hadoop.util.Tool;
import org.apache.hadoop.util.ToolRunner;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Arrays;

public class Finalizer extends Configured implements Tool
{
	private final static Log LOG = LogFactory.getLog(Finalizer.class);

	public static void main(String[] args) throws Exception
	{
		int ret = ToolRunner.run(new Finalizer(), args);
		System.exit(ret);

	}

	public int run(String[] arg0) throws Exception
	{
		// create a dummy job to get a loaded hadoop configuration
		Job j = new Job();
		Configuration config = j.getConfiguration();
		KrakeConfig.load();

		FileSystem fs = FileSystem.get(config);

		LOG.info("Finalizing graph...");
		finalizeWebGraph(fs,
				new Path(KrakeConfig.transient_webgraph),
				new Path(KrakeConfig.final_webgraph),
				new Path(KrakeConfig.id_to_hash_path));
		LOG.info("done");
		return 0;
	}
	
	private static void finalizeWebGraph(FileSystem fs, Path graphIn, Path graphOut, Path idToHash) throws IOException
	{
		FSDataInputStream in = fs.open(graphIn);
		FSDataOutputStream out = fs.create(graphOut);
		
		// read the mapping to get the expected node count in the graph (silly)
		URLMapping mapping = URLMapping.openDFS(fs, idToHash);
		int node_count = mapping.getSize();
		mapping = null;
		
		out.writeInt(node_count);
		BufferedReader r = new BufferedReader(new InputStreamReader(in));
		
		int processed_nodes = 0;
		while(true)
		{
			String line = r.readLine();
			if(line == null)
			{
				if(processed_nodes != node_count)
					throw new IOException("Unexpected EOF");
				break;
			}
			
			String[] segments = line.split("\\s+");
			
			if(Integer.parseInt(segments[0]) - 1 != processed_nodes)
				throw new IOException("Invalid graph file, nodes are not sorted or skip numbers");
			
			int inlink_count = 0;
			int[] links = new int[segments.length - 1];
			for(int i = 1; i < segments.length; ++i)
			{
				if(segments[i].length() == 0)
				{
					LOG.warn("Empty string in line segments, format corrupted?");
					continue;
				}
				
				links[i-1] = Integer.parseInt(segments[i]);
				
				// count inlinks
				if(links[i-1] < 0)
					++inlink_count;
			}
			
			// sort will put inlinks (negative) first followed by the outlinks (positive)
			Arrays.sort(links);
			
			out.writeInt(links.length + 1); // ints to read for this node
			out.writeInt(inlink_count);
			
			for(int link : links)
				out.writeInt(Math.abs(link) - 1); // remove sign and bias before writing
			
			++processed_nodes;
			
			if(processed_nodes % 1000 == 0)
				LOG.info("Written " + processed_nodes + " nodes");
		}
		
		in.close();
		out.close();
	}

	
	@SuppressWarnings("unused")
	private static void finalizeIdToHashMapping(FileSystem fs, Path hashIn, Path hashOut) throws IOException
	{
		FSDataInputStream in = fs.open(hashIn);
		FSDataOutputStream out = fs.create(hashOut);
		
		long size = fs.getFileStatus(hashIn).getLen();
		long elements = size / (4 + 20);
		if(size % elements != 0)
			throw new IOException("Corrupt id to hash mapping");
		
		out.writeInt((int)elements);
		LOG.info(elements + " elements in mapping");
		
		int id = -1;
		byte[] hash = new byte[20];
		for(long i = 0; i < elements; ++i)
		{
			int cid = in.readInt();
			if(id + 1 != cid)
				throw new IOException("Invalid id sequence in hash mapping");

			id = cid;
			in.readFully(hash);
			out.write(hash);
		}
		
		in.close();
		out.close();
	}
/*
	private static void finalizeIdToURLMapping(FileSystem fs, Path mappingIn, Path mappingOut, Path idToHash) throws IOException
	{
		FSDataInputStream in = fs.open(mappingIn);
		FSDataOutputStream out = fs.create(mappingOut);

		URLMapping hashToId = URLMapping.openDFS(fs, idToHash);

		BufferedReader bread = new BufferedReader(new InputStreamReader(in, Charset.forName("UTF-16")));
		
		ByteArrayOutputStream urlbuffer = new ByteArrayOutputStream((int)fs.getFileStatus(mappingIn).getLen());
		ByteArrayOutputStream intbuffer = new ByteArrayOutputStream(hashToId.getSize() * 2 * 4 + 1);
		DataOutputStream intstream = new DataOutputStream(intbuffer);
		
		out.writeInt(hashToId.getSize());
		int offset = 0;
		int expected_id = 0;
		
		while(true)
		{
			String line = bread.readLine();
			if(line == null)
				break;
			String[] elements = line.split("\\t");
			if(elements.length != 2)
				throw new IOException("Invalid format in mapping");
			
			String shash = elements[0];
			byte[] hash = Common.HextoByteArray(shash);
			int real_id = hashToId.hash2int(hash);
			if(expected_id != real_id)
				throw new IOException("Mappings are out of sync");
			++expected_id;
			
			byte[] urlbytes = elements[1].getBytes();
			urlbuffer.write(urlbytes);
			int written = urlbytes.length;
			
			intstream.writeInt(offset);
			intstream.writeInt(written);
			offset += written;
		}
		
		// free as many resources as we can
		bread.close();
		in.close();
		hashToId = null;
		System.gc();
		
		out.write(intbuffer.toByteArray());
		
		// free as many resources as we can
		intstream.close();
		intstream = null;
		intbuffer = null;
		System.gc();
		
		byte[] urls = urlbuffer.toByteArray();
		
		// free as many resources as we can
		urlbuffer = null;
		System.gc();
		
		out.write(urls);
		out.close();
	}
 */

}
