/*
Krake Webcrawlerframework
Copyright (C) LS2 TU Dortmund (http://ls2-www.cs.tu-dortmund.de)

This file is part of the Krake Webcrawlerframework.
The sourcecode and data files provided come without any warranty. Use at your own risk!

The sourcecode is the intellectual property of its authors. You have permission to download
and use this sourcecode for the sake of inspection and for private and academic use. You are
not allowed to redistribute this sourcecode without our permission. If you want to use the
sourcecode in any other way as granted above,
please contact Christian Sohler <christian.sohler[at]tu-dortmund.de>.
*/
package krake;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.yaml.snakeyaml.Yaml;

import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;

/**
 * Holds global configuration settings for all krake-classes and processes.
 * Must be loaded before any other krake-class is used.
 * If not specified otherwise, the location of the configfile is read from the environment variable KRAKE_CONFIG_PATH
 * @author Thomas Nitschke
 */
public class KrakeConfig
{
	private final static Log LOG = LogFactory.getLog(KrakeConfig.class);
	@SuppressWarnings("rawtypes")
	private static Map parsed = null;

	/**
	 * Loads the config from the local path in the environment variable KRAKE_CONFIG_PATH
	 * @throws IOException
	 */
	public static void load() throws IOException
	{
		if(isLoaded())
			return;

		String path = System.getenv("KRAKE_CONFIG_PATH");
		if(path == null)
			throw new IOException("KRAKE_CONFIG_PATH not specified in env");

		load(path);
	}

	/**
	 * Loads the configfile from the specified local path.
	 * @param file Path to the local configfile
	 * @throws IOException
	 */
	public static void load(String file) throws IOException
	{
		try
		{
			loadUnsafe(file);
		}
		catch(Exception e)
		{
			parsed = null;

			LOG.error("Invalid config file", e);
			throw new IOException(e);
		}
	}

	/**
	 * Returns true if the config has been loaded successfully.
	 */
	public static boolean isLoaded()
	{
		return parsed != null;
	}

	@SuppressWarnings({ "rawtypes", "unchecked" })
	private static void loadUnsafe(String file) throws FileNotFoundException
	{
		if(isLoaded())
			return;

		Yaml yaml = new Yaml();
		parsed = (Map)yaml.load(new FileReader(file));

		Map key_storage = (Map)parsed.get("key_storage");
		key_backend = (String)key_storage.get("backend");
		key_replication = (Integer)key_storage.get("replication");

		// load server list
		List _servers = (List)parsed.get("servers");
		servers = new String[_servers.size()];
		_servers.toArray(servers);

		// mysql settings
		if(key_backend.equals("mysql"))
		{
			Map mysql = (Map)key_storage.get("mysql");


			mysql_db = (String)mysql.get("db");
			mysql_user = (String)mysql.get("user");
			mysql_passwd = (String)mysql.get("passwd");
			mysql_port = (Integer)mysql.get("port");
		}

		// crawl options
		Map crawl = (Map)parsed.get("crawl");
		crawl_max_site_size = (Integer)crawl.get("max_site_size");
		crawl_timeout = (Integer)crawl.get("timeout");
		crawl_user_agent = (String)crawl.get("user_agent");
		crawl_base = (String)crawl.get("crawl_base");
		batch_size = (Integer)crawl.get("batch_size");


		// export options
		Map export = (Map)parsed.get("export");
		content_path = (String)export.get("content_path");
		offsets_path = (String)export.get("offsets_path");

		id_to_hash_path = (String)export.get("id_to_hash_path");
		hash_to_url_path = (String)export.get("hash_to_url_path");

		transient_webgraph = (String)export.get("transient_webgraph");
		final_webgraph = (String)export.get("final_webgraph");

		// queue options
		Map queue = (Map)parsed.get("queue");
		queue_backend = (String)queue.get("backend");
		queue_local_dir = (String)queue.get("local_dir");

		// load filters
		List patterns = (List)parsed.get("filters");
		ArrayList<Pattern> tmp = new ArrayList<Pattern>(patterns.size());
		for(int i = 0; i < patterns.size(); ++i)
		{
			String pattern_string = (String)patterns.get(i);
			try
			{
				tmp.add(Pattern.compile(pattern_string));
			}
			catch(PatternSyntaxException e)
			{
				LOG.warn("Unable to load filtering-pattern '" + pattern_string + "'", e);
			}
		}

		filters = new Pattern[tmp.size()];
		tmp.toArray(filters);
	}


	// See the sample configfile or manual for a detailed description of these variables.
	public static String[] servers = null;
	public static int batch_size;

	public static String key_backend;
	public static int key_replication;
	public static int mysql_port;
	public static String mysql_user;
	public static String mysql_passwd;
	public static String mysql_db;

	public static int crawl_max_site_size = 1024*1024*4; // 4 MB
	public static String crawl_user_agent = "TU Dortmund CS-LS2 Krakeprojekt Crawler";
	public static int crawl_timeout = 5000;
	public static String crawl_domainhistogram_path = null;
	public static String crawl_domainfilter_path = null;
	public static String crawl_base = null;


	public static String content_path = null;
	public static String offsets_path = null;
	public static String id_to_hash_path = null;
	public static String hash_to_url_path = null;
	public static String transient_webgraph = null;
	public static String final_webgraph = null;

	public static String queue_local_dir;
	public static String queue_backend;
	public static Pattern[] filters = new Pattern[0];
}
