/*
Krake Webcrawlerframework
Copyright (C) LS2 TU Dortmund (http://ls2-www.cs.tu-dortmund.de)

This file is part of the Krake Webcrawlerframework.
The sourcecode and data files provided come without any warranty. Use at your own risk!

The sourcecode is the intellectual property of its authors. You have permission to download
and use this sourcecode for the sake of inspection and for private and academic use. You are
not allowed to redistribute this sourcecode without our permission. If you want to use the
sourcecode in any other way as granted above,
please contact Christian Sohler <christian.sohler[at]tu-dortmund.de>.
*/
package krake.crawler;

import krake.Common;
import krake.KrakeConfig;
import krake.keys.KeyInfo;
import krake.keys.KeyStorage;
import krake.mapreduce.CrawlMapper;
import krake.queue.Queue;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapreduce.Counter;

import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.net.*;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Main crawlerclass, contains the download- and processinglogic for webpages.
 * 
 * @author Alex Munteanu, Thomas Nitschke, Marc Gille
 */
public class KrakeCrawler
{
	final static Log LOG = LogFactory.getLog(KrakeCrawler.class);

	/**
	 * Interface to the key-storage
	 */
	private KeyStorage m_keys = null;

	/**
	 * Interface to the queue
	 */
	private Queue m_queue = null;

	/**
	 * Manages robots.txt files
	 */
	private RobotsManager m_robots = new RobotsManager();

	/**
	 * Number of currently processed URLs
	 */
	private int m_local_input_urls = 0;

	/**
	 * Cache for recently discovered URLs
	 */
	private URLCache m_isseen_cache = new URLCache();

	/**
	 * Hadoop-Context to emit key/value-pairs
	 */
	private krake.mapreduce.CrawlMapper.Context m_context = null;

	// Hadoop counters for statistics
	private Counter ctr_robots;
	private Counter ctr_filter;
	private Counter ctr_error;
	private Counter ctr_size;
	private Counter ctr_header;
	private Counter ctr_type;
	private Counter ctr_timeout;
	private Counter ctr_new;
	private Counter ctr_seen;
	private Counter ctr_input;
	private Counter ctr_redirect;
	private Counter ctr_invalid;

	/**
	 * Downloads the content of the UrlConnections related URL and tries to
	 * decode the contents with the specified encoding.
	 *
	 * @param urlConnection
	 * 			Connection to use
	 * @param encoding
	 *            Encodings (UTF-8, Latin1 etc)
	 * @return Contents of the URL
	 * @throws SiteTooLargeException
	 *             Content is bigger than Common.MAX_SITE_SIZE
	 * @throws java.io.IOException
	 *             Other IO-Error
	 */
	private static String getPageContent(URLConnection urlConnection, String encoding) throws CrawlException, IOException
	{
		StringBuilder buff = new StringBuilder(8192);

		Reader r = new InputStreamReader(urlConnection.getInputStream(), encoding);

		char b[] = new char[1024];
		int numRead = r.read(b);
		int totalRead = 0;
		while(numRead != -1)
		{
			totalRead += numRead;
			if(totalRead >= KrakeConfig.crawl_max_site_size)
				throw new CrawlException(KeyInfo.ERROR_SITE_TOO_BIG, urlConnection.getURL().toString());

			buff.append(b, 0, numRead);
			numRead = r.read(b);
		}
		r.close();
		return buff.toString();
	}

	/**
	 * Intializes the crawler for this mapper instance. Every crawler instance
	 * needs one mapper instance.
	 * 
	 * @param ctx
	 *            Context of the hadoop mapper
	 * @throws IOException
	 */
	public void init(CrawlMapper.Context ctx) throws IOException
	{
		m_context = ctx;

		// set default for URL access
		URLConnection.setDefaultAllowUserInteraction(false);

		// connect to keystorage
		m_keys = KeyStorage.createFromConfig();

		// get a handle of the queue
		m_queue = Queue.createFromConfig();
		m_queue.preCrawl();

		// init hadoop counters
		ctr_robots = ctx.getCounter(Common.CrawlCounters.BLOCKED_BY_ROBOTS);
		ctr_filter = ctx.getCounter(Common.CrawlCounters.BLOCKED_BY_FILTER);
		ctr_error = ctx.getCounter(Common.CrawlCounters.BLOCKED_BY_ERROR);
		ctr_size = ctx.getCounter(Common.CrawlCounters.BLOCKED_BY_SIZE);
		ctr_type = ctx.getCounter(Common.CrawlCounters.BLOCKED_BY_TYPE);
		ctr_header = ctx.getCounter(Common.CrawlCounters.BLOCKED_BY_HEADER);
		ctr_timeout = ctx.getCounter(Common.CrawlCounters.BLOCKED_BY_TIMEOUT);
		ctr_new = ctx.getCounter(Common.CrawlCounters.NEW_URLS);
		ctr_seen = ctx.getCounter(Common.CrawlCounters.SEEN_URLS);
		ctr_input = ctx.getCounter(Common.CrawlCounters.INPUT_URLS);
		ctr_redirect = ctx.getCounter(Common.CrawlCounters.BLOCKED_BY_REDIRECT);
		ctr_invalid = ctx.getCounter(Common.CrawlCounters.BLOCKED_BY_INVALID_URL);
	}


	/**
	 * Free remaining resources. Must be called once after the last crawling
	 * operation.
	 */
	public void shutDown()
	{
		try
		{
			m_keys.close();
		}
		catch(Exception e)
		{
			LOG.error(e);
		}

		try
		{
			// reorganize the queue
			m_queue.postCrawl(m_local_input_urls);
		}
		catch(Exception e)
		{
			LOG.error(e);
		}
	}

	/**
	 * Crawls, processes and stores the given URL
	 * 
	 * @param surl
	 * @throws IOException
	 */
	public void processURL(String hash, String surl) throws IOException
	{
		ctr_input.increment(1);
		++m_local_input_urls;

		URL url = new URL(surl);

		// simple sanity check
		if(url.getProtocol() == null || url.getHost() == null || url.getPath() == null || url.getHost().equals(""))
		{
			ctr_invalid.increment(1);
			return;
		}

		try
		{
			if(isFiltered(url))
				return;

			HttpURLConnection connection = (HttpURLConnection)url.openConnection();

			// set behaviour properties
			connection.setRequestProperty("User-Agent", KrakeConfig.crawl_user_agent);
			connection.setConnectTimeout(KrakeConfig.crawl_timeout);
			connection.setReadTimeout(KrakeConfig.crawl_timeout);
			connection.setInstanceFollowRedirects(true); // java aborts after 20 redirects
			connection.setAllowUserInteraction(false);

			crawlURL(hash, url, connection);
			connection.disconnect();
			connection = null;
		}
		catch(ClassCastException e)
		{
			LOG.error(e);
			ctr_error.increment(1);
			throw new IOException(e);// we should really know about that
		}
		catch(ProtocolException e)
		{
			markAsErroneousURL(url, hash, KeyInfo.ERROR_PROTOCOL);

			// invalid redirect
			if(e.getMessage().indexOf("redirected") != -1)
				ctr_redirect.increment(1);
			else // other error
			{
				LOG.error(e);
				ctr_error.increment(1);
			}
		}
		catch(SocketTimeoutException e)
		{
			markAsErroneousURL(url, hash, KeyInfo.ERROR_TIMEOUT);
			LOG.error(e);
			ctr_timeout.increment(1);
		}
		catch(CrawlException e)
		{
			markAsErroneousURL(url, hash, e.getErrorCode());
			LOG.error(e);
			ctr_error.increment(1);
		}
		catch(Exception e)
		{
			markAsErroneousURL(url, hash, KeyInfo.ERROR_GENERIC);
			LOG.error(e);
			ctr_error.increment(1);
		}
	}

	/**
	 * Returns true if the url should not be crawled
	 * 
	 * @param url
	 * @return
	 */
	private boolean isFiltered(URL url)
	{
		// Can only search http: protocol URLs
		if(!Common.isSupportedProtocol(url.getProtocol()) || Common.isNonTextURL(url))
		{
			ctr_filter.increment(1);
			return true;
		}

		// test each regexp pattern against the url
		String surl = url.toString();
		for(Pattern p : KrakeConfig.filters)
		{
			Matcher m = p.matcher(surl);
			if(m.find())
				return true;
		}

		// Make sure it is safe to crawl
		if(!m_robots.isSafe(url))
		{
			ctr_robots.increment(1);
			return true;
		}

		return false;
	}

	/**
	 * Stores an errorcode for the given URL in the KeyDB.
	 * 
	 * @param url
	 * @param ec
	 */
	private void markAsErroneousURL(URL url, String hash, KeyInfo ec)
	{
		try
		{
			m_keys.setKey(hash, ec);
		}
		catch(IOException e)
		{
			LOG.error(e);
		}
	}


	/**
	 * Add the URL to the queue
	 * 
	 * @param url
	 * @throws IOException
	 */
	protected void queueURL(URL url) throws IOException
	{
		m_queue.add(url.toString());
	}

	/**
	 * Add the URL to the seen-table
	 * 
	 * @param url
	 * @throws IOException
	 */
	protected void addSeenURL(URL url, String hash) throws IOException
	{
		m_keys.setKey(hash, KeyInfo.SEEN);
	}

	/**
	 * Check if the URL is in the seen-table
	 * 
	 * @param url
	 * @return
	 * @throws IOException
	 */
	protected boolean isKnownURL(URL url, String hash) throws IOException
	{
		// ask the local cache first
		if(!m_isseen_cache.query(url))
		{
			// ask the db filter
			try
			{
				return m_keys.isKnown(hash);
			}
			catch(Exception e)
			{
				e.printStackTrace();
				throw new IOException("isSeen query failed!");
			}
		}
		return true;
	}

	/**
	 * Crawls the URL
	 * 
	 * @param url
	 * @throws IOException On everything the CrawlException didn't handle
	 * @throws InterruptedException
	 * @throws CrawlException On error-response, timeouts, invalid data etc.
	 */
	protected void crawlURL(String hash, URL url, HttpURLConnection urlConnection) throws IOException, InterruptedException, CrawlException
	{
		// download
		String content = downloadContent(urlConnection);

		// write/queue links
		processLinksInContent(url, content);

		// write content
		m_context.write(new Text(url.toString()), new Text(content));

		// mark as crawled
		m_keys.setKey(hash, KeyInfo.CRAWLED);
	}

	private String downloadContent(HttpURLConnection urlConnection) throws IOException, CrawlException
	{
		// Only access pages returning 200 OK status
		if(urlConnection.getResponseCode() != HttpURLConnection.HTTP_OK)
		{
			ctr_header.increment(1);
			throw new CrawlException(KeyInfo.ERROR_NON200_RESPONSE);
		}

		// parse the content-length http field if available
		int content_size = urlConnection.getContentLength();
		if(content_size >= KrakeConfig.crawl_max_site_size)
		{
			ctr_size.increment(1);
			throw new CrawlException(KeyInfo.ERROR_SITE_TOO_BIG);
		}

		// parse type and encoding
		String type = urlConnection.getContentType();
		if(type == null)
		{
			ctr_type.increment(1);
			throw new CrawlException(KeyInfo.ERROR_INVALID_CONTENT_TYPE);
		}

		String encoding = urlConnection.getContentEncoding();
		if(encoding == null)
			encoding = "ISO8859_1";

		String[] fields = type.split(";");
		boolean is_texthtml = false;
		for(int i = 0; i < fields.length; ++i)
		{
			fields[i] = fields[i].trim();
			if(fields[i].equals("text/html"))
				is_texthtml = true;
		}

		if(!is_texthtml)
		{
			ctr_error.increment(1);
			throw new CrawlException(KeyInfo.ERROR_CONTENT_IS_NON_HTML);
		}

		String content;
		try
		{
			content = getPageContent(urlConnection, encoding);
		}
		catch(CrawlException e)
		{
			if(e.getErrorCode() == KeyInfo.ERROR_SITE_TOO_BIG)
				ctr_size.increment(1);
			throw e;
		}
		return content;
	}

	/**
	 * Reads cookie data out of the metadata of the http connection. The
	 * cookiedata is combined into a single string with a linefeed after each
	 * cookie.
	 * 
	 * @param urlConnection
	 * @return
	 */
	@SuppressWarnings("unused")
	private String extractCookies(URLConnection urlConnection)
	{
		List<String> cookies_rfc2109 = urlConnection.getHeaderFields().get("Set-Cookie");
		List<String> cookies_rfc2965 = urlConnection.getHeaderFields().get("Set-Cookie2");

		StringBuilder buff = new StringBuilder(1024);

		if(cookies_rfc2109 != null)
		{
			for(String s : cookies_rfc2109)
				buff.append(s);
		}

		if(cookies_rfc2965 != null)
		{
			for(String s : cookies_rfc2965)
				buff.append(s);
		}

		return buff.toString();
	}

	/**
	 * Extract all URLs from the supplied text and count the additional
	 * outlinks.
	 * 
	 * @param url
	 *            Origin of the content
	 * @param content
	 *            Text of the URL
	 * @return List of URLs and the number of additional outlinks.
	 * @throws IOException
	 */
	private void processLinksInContent(URL url, String content) throws IOException
	{
		String lowerCaseContent = content.toLowerCase();
		Set<String> outlinks = new HashSet<String>(128);
		int new_urls = 0;
		int seen_urls = 0;

		// extract links in hrefs and frames
		extractURLsByPattern(url, content, lowerCaseContent, Common.LinkPattern, outlinks);
		extractURLsByPattern(url, content, lowerCaseContent, Common.FramePattern, outlinks);

		// run over the collected links and add the unknown ones to the queue
		for(String surl : outlinks)
		{
			// urlparsing should not fail, (already done in extractURLsByPattern)
			URL link_url = new URL(surl);
			String hash = Common.normalize(link_url);

			// don't add filtered stuff to the queue
			if(isFiltered(link_url))
				continue;

			// is unknown
			if(!isKnownURL(link_url, hash))
			{
				addSeenURL(link_url, hash);
				queueURL(link_url);
				++new_urls;

				// enhanced crawling logic: always queue a link to the homepage TODO make this optional in config?
				if(extractAndQueueHomepage(link_url))
					++new_urls;
			}
			else
				++seen_urls;
		}

		ctr_new.increment(new_urls);
		ctr_seen.increment(seen_urls);
	}

	/**
	 * Extract all URLs from the supplied text with a specific pattern.
	 * 
	 * @param url
	 *            Origin of the content
	 * @param content
	 *            Text of the URL
	 * @param lowerCaseContent
	 *            Text of the URL, converted to lower case
	 * @param p
	 *            Pattern used for matching
	 * @param outlinks
	 *            Set to which found links will be added
	 * @throws IOException
	 */
	private void extractURLsByPattern(URL url, String content, String lowerCaseContent, Pattern p, Set<String> outlinks) throws IOException
	{
		Matcher m = p.matcher(lowerCaseContent);
		int start = 0;
		String href;
		while(m.find(start))
		{
			href = null;
			start = m.end();

			// should never happen since the pattern matched
			if(m.groupCount() != 2)
			{
				System.out.println("Broken Pattern");
				continue;
			}

			// get the link from the original content
			// since the lowercase conversion destroys paths
			if(m.start(1) != -1 && m.end(1) != -1)
				href = content.substring(m.start(1), m.end(1)).trim();
			else if(m.start(2) != -1 && m.end(2) != -1)
				href = content.substring(m.start(2), m.end(2)).trim();

			// early skip for links that we don't want
			if(href.toLowerCase().startsWith("javascript") || href.startsWith("#"))
				continue;

			URL urlLink;
			try
			{
				urlLink = new URL(url, href);
				href = urlLink.toString();
				if(href.indexOf("#") != -1)
					href = href.substring(0, href.indexOf("#"));

				int quoteidx = href.indexOf("'");
				int dblquoteidx = href.indexOf("\"");
				quoteidx = (quoteidx == -1) ? href.length() : quoteidx;
				dblquoteidx = (dblquoteidx == -1) ? href.length() : dblquoteidx;

				if(quoteidx < dblquoteidx)
					href = href.substring(0, quoteidx);
				else
					href = href.substring(0, dblquoteidx);

				urlLink = new URL(url, href);
				href = urlLink.toString(); // final combined and stripped url
			}
			catch(MalformedURLException e)
			{
				continue;
			}

			// add link
			outlinks.add(href);

			// offset by match
			start = m.end();
		}
	}

	/**
	 * Extracts the homepage-url of a given url (www.bla.com/foo/bar ->
	 * www.bla.com) and adds it to the queue if it isn't already known.
	 * 
	 * @param urlLink
	 * @return true if the homepage-url was added to the queue, false otherwise
	 * @throws IOException
	 */
	private boolean extractAndQueueHomepage(URL urlLink) throws IOException
	{
		// extract the domain and add it as homepage link
		try
		{
			URL homeURL = new URL(urlLink.getProtocol(), urlLink.getHost(), "");
			String hash = Common.normalize(homeURL);
			if(!isKnownURL(homeURL, hash))
			{
				addSeenURL(homeURL, hash);
				queueURL(homeURL);
				return true;
			}
		}
		catch(MalformedURLException e)
		{
			// not really a servere error
			LOG.warn(e);
		}

		return false;
	}
}
