/*
Krake Webcrawlerframework
Copyright (C) LS2 TU Dortmund (http://ls2-www.cs.tu-dortmund.de)

This file is part of the Krake Webcrawlerframework.
The sourcecode and data files provided come without any warranty. Use at your own risk!

The sourcecode is the intellectual property of its authors. You have permission to download
and use this sourcecode for the sake of inspection and for private and academic use. You are
not allowed to redistribute this sourcecode without our permission. If you want to use the
sourcecode in any other way as granted above,
please contact Christian Sohler <christian.sohler[at]tu-dortmund.de>.
*/
package krake.crawler;

import krake.KrakeConfig;

import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.util.HashMap;
import java.util.StringTokenizer;

/**
 * Best effort (read: quick & dirt) implementation of a robots.txt parser and manager.
 * @author Alex Munteanu
 */
public class RobotsManager
{
	/**
	 * Max. amount of entries in the robotCache
	 */
	private final int MAX_ROBOTS = 100;

	/**
	 * Randomized cache for robotentries
	 */
	private HashMap<String, String> robotCache = new HashMap<String, String>(150);

	/**
	 * Determine if this URL is safe to read from robots.txt
	 *
	 * @param url
	 * @return indicates whether the path is allowed to be crawled
	 */
	public boolean isSafe(URL url)
	{
		String strCommands;
		String strHost = url.getHost();

		if(!robotCache.containsKey(strHost.toLowerCase()))
		{
			// form URL of the robots.txt file
			String strRobot = "http://" + strHost + "/robots.txt";
			URL urlRobot;

			try
			{
				urlRobot = new URL(strRobot);
			}
			catch(MalformedURLException e)
			{
				// something weird is happening, so don't trust it
				return false;
			}

			try
			{
				URLConnection con = urlRobot.openConnection();
				con.setConnectTimeout(KrakeConfig.crawl_timeout);
				con.setReadTimeout(KrakeConfig.crawl_timeout);
				con.setAllowUserInteraction(false);
				con.setRequestProperty("User-Agent", KrakeConfig.crawl_user_agent);

				if(((HttpURLConnection) con).getResponseCode() != HttpURLConnection.HTTP_OK)
				{
					if(robotCache.size() >= MAX_ROBOTS)
						randomDelete(1);
					robotCache.put(strHost.toLowerCase(), "");
					return true;
				}

				StringBuilder buff = new StringBuilder(1024);
				String enc = con.getContentEncoding();// FIXME: this may fire an
				// IllegalArgumentException
				// when the webserver is very weird (like ngx with incorrect
				// gzip compression)
				if(enc == null)
					enc = "ISO8859_1";
				Reader r = new InputStreamReader(con.getInputStream(), enc);

				char b[] = new char[1024];
				int numRead = r.read(b);
				while(numRead != -1)
				{
					buff.append(b, 0, numRead);
					numRead = r.read(b);
				}
				r.close();

				strCommands = buff.toString();
			}
			catch(IOException e)
			{
				// if there is no robots.txt file, it is OK to search
				if(robotCache.size() >= MAX_ROBOTS)
					randomDelete(1);
				robotCache.put(strHost.toLowerCase(), "");
				return true;
			}
			catch(IllegalArgumentException e)
			{
				// if there is no robots.txt file, it is OK to search
				if(robotCache.size() >= MAX_ROBOTS)
					randomDelete(1);
				robotCache.put(strHost.toLowerCase(), "");
				return true;
			}
			if(robotCache.size() >= MAX_ROBOTS)
				randomDelete(1);
			robotCache.put(strHost.toLowerCase(), strCommands);
		}
		else
			strCommands = robotCache.get(strHost.toLowerCase());

		if(strCommands == null || strCommands.equals(""))
			return true;
		strCommands = strCommands.toLowerCase();// tolower case since some sites use User-Agent/User-agent etc.

		// assume that this robots.txt refers to us and
		// search for "Disallow:" commands.
		String strURL = url.getFile();
		int index = 0;
		int endindex = 0;
		index = strCommands.indexOf("user-agent: *", index);
		endindex = strCommands.indexOf("user-agent:", index + 1);
		if(index == -1)
			index = Integer.MAX_VALUE;
		if(endindex == -1)
			endindex = Integer.MAX_VALUE;

		while((index = strCommands.indexOf("disallow:", index)) != -1 && index < endindex)
		{
			index += "disallow:".length();
			String strPath = strCommands.substring(index);
			StringTokenizer st = new StringTokenizer(strPath);

			if(!st.hasMoreTokens())
				break;

			String strBadPath = st.nextToken().replaceAll("\\*", "");

			// if the URL starts with a disallowed path, it is not safe
			if(strURL.indexOf(strBadPath) == 0)
				return false;
		}

		return true;
	}


	/**
	 * Randomly deletes the given number of entries in the robot cache.
	 *
	 * @param cnt
	 *            The number of entries to delete.
	 */
	private void randomDelete(int cnt)
	{
		if(cnt >= robotCache.size())
			robotCache.clear();
		else
		{
			for(int i = 0; i < cnt; ++i)
			{
				String[] keys = robotCache.keySet().toArray(new String[0]);

				robotCache.remove(keys[(int) (Math.random() * keys.length)]);
			}
		}
	}
}
