/*
Krake Webcrawlerframework
Copyright (C) LS2 TU Dortmund (http://ls2-www.cs.tu-dortmund.de)

This file is part of the Krake Webcrawlerframework.
The sourcecode and data files provided come without any warranty. Use at your own risk!

The sourcecode is the intellectual property of its authors. You have permission to download
and use this sourcecode for the sake of inspection and for private and academic use. You are
not allowed to redistribute this sourcecode without our permission. If you want to use the
sourcecode in any other way as granted above,
please contact Christian Sohler <christian.sohler[at]tu-dortmund.de>.
*/
package krake.file;

import krake.Common;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FSDataInputStream;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;

import java.io.EOFException;
import java.io.IOException;


/**
 * Provides random access to a krake content file in the hadoop DFS.
 * @author Thomas Nitschke
 */
public class ContentDFSRandomAccess extends ContentRandomAccess
{
	private static final Log LOG = LogFactory.getLog(ContentDFSRandomAccess.class);
	private FSDataInputStream m_in;

	/**
	 * Loads the offsetfile and opens the content file for random access. The entire offsetfile
	 * will be loaded into memory!
	 * @param config Hadoop config for the cluster, must contain all DFS related settings
	 * @param offsetfile DFS path to the offset file
	 * @param contentfile DFS path to the content file
	 * @throws IOException
	 * @author Thomas Nitschke
	 */
	public ContentDFSRandomAccess(Configuration config, Path offsetfile, Path contentfile) throws IOException
	{
		FileSystem dfs = FileSystem.get(config);
		FSDataInputStream offstream = dfs.open(offsetfile);
		long length = dfs.getFileStatus(offsetfile).getLen();
		readOffsetTable(offstream, length);
		offstream.close();
		
		LOG.info("Opening content file");
		m_in = dfs.open(contentfile);
		
		LOG.info("Verifying");
		verify();
		m_ids = null;
	}
	
	@Override
	public void close()
	{
		try
		{
			m_in.close();
		}
		catch(IOException e)
		{
			e.printStackTrace();
		}
	}
	
	private int readSize(FSDataInputStream in) throws IOException
	{
		byte[] integer = new byte[4];
		readBuffer(in, integer, integer.length);
		return Common.intFromBytes(integer);
	}

	private void readBuffer(FSDataInputStream in, byte[] buffer, int to_read) throws IOException
	{
		int off = 0;
		int missing = to_read;
		int read;

		while(missing > 0)
		{
			read = in.read(buffer, off, missing);
			if(read == -1)
				throw new EOFException();

			off += read;
			missing -= read;
		}
	}

	@Override
	public String get(int id) throws IOException
	{
		if(id >= getEntryCount())
			return null;
		
		long offset = m_offsets[id];
		
		m_in.seek(offset);
		int read_size = readSize(m_in);
		
		if(read_size == 0)
			return "";
		
		byte[] buffer = new byte[read_size];
		readBuffer(m_in, buffer, buffer.length);
		
		return uncompress(buffer);
	}

	@Override
	public String get_next() throws EOFException, IOException
	{
		// read the size of the last node
		readSize(m_in);
		
		// read the size of the new node
		int read_size = readSize(m_in);
		if(read_size == 0)
			return "";
		
		byte[] buffer = new byte[read_size];
		readBuffer(m_in, buffer, buffer.length);
		
		return uncompress(buffer);
	}
}
