/*
Krake Webcrawlerframework
Copyright (C) LS2 TU Dortmund (http://ls2-www.cs.tu-dortmund.de)

This file is part of the Krake Webcrawlerframework.
The sourcecode and data files provided come without any warranty. Use at your own risk!

The sourcecode is the intellectual property of its authors. You have permission to download
and use this sourcecode for the sake of inspection and for private and academic use. You are
not allowed to redistribute this sourcecode without our permission. If you want to use the
sourcecode in any other way as granted above,
please contact Christian Sohler <christian.sohler[at]tu-dortmund.de>.
*/
package krake.file;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import java.io.DataInputStream;
import java.io.EOFException;
import java.io.IOException;
import java.io.InputStream;

/**
 * Basis for all classes that access the krake content file as a random access storage.
 * In order to provide random access an offset table is read into memory and used for seeks/jumps
 * in the content file.
 * @author Thomas Nitschke
 */
public abstract class ContentRandomAccess extends Content
{
	private final Log LOG = LogFactory.getLog(ContentRandomAccess.class);
	protected int[] m_ids = null;
	protected long[] m_offsets = null;
	private int m_count = 0;

	/**
	 * Reads the offset table into memory.
	 * @param in
	 * @param filesize
	 * @throws IOException
	 */
	protected void readOffsetTable(InputStream in, long filesize) throws IOException
	{
		int entries = (int)(filesize / (4+8)); // one int and long for each entry
		
		LOG.info("Reading " + filesize + " bytes of offsettable with " + entries + " entries");
		
		m_ids = new int[entries];
		m_offsets = new long[entries];
		
		DataInputStream din = new DataInputStream(in);
		for(int i = 0; i < entries; ++i)
		{
			m_ids[i] = din.readInt();
			if(i == 0)
				LOG.info("First id is " + m_ids[i]);
			m_offsets[i] = din.readLong();
		}
		
		m_count = m_ids.length;
		LOG.info("Done");
	}

	/**
	 * Retuns number of entries in the content file.
	 */
	public int getEntryCount()
	{
		return m_count;
	}

	/**
	 * Checks the content file and offset table for consistency:
	 * @throws IOException
	 */
	protected void verify() throws IOException
	{
		int last_id = -1;
		long last = -1;

		// ids must start at 0
		if(m_ids[0] != 0)
			throw new IOException("Ids don't start at 0: " + m_ids[0]);
		
		for(int i = 0; i < m_ids.length; ++i)
		{
			// ids must increase by one for each entry
			if(m_ids[i] < 0 || m_ids[i] - last_id != 1)
			{
				System.out.println("Invalid id at " + i + " " + m_ids[i] + " last " + last_id);
				// HACK FIXME
				m_count = i;
				return;
				
				//throw new IOException("Corrupted offset-file, id invalid. Delta greater one.");
			}

			// offsets must be positive and must not overlap
			if(m_offsets[i] < 0 || m_offsets[i] <= last)
				throw new IOException("Corrupted offset-file, offset invalid. Offset is out of bounds.");
			
			last = m_offsets[i];
			last_id = m_ids[i];
		}
	}
	
	/**
	 * Returns the content of a node or an empty string if the node doesn't have any content.
	 * @param id
	 * @return Contentstring or ""
	 * @throws IOException If id is out of bounds
	 */
	public abstract String get(int id) throws IOException;
	
	/**
	 * Returns the content of the successor-node of the last get() call. Example: After calling get(12), get_next() would return node 13.
	 * get() must be called at least once before calling this function to initialize the interal file pointers.
	 * @return Contentstring or ""
	 * @throws IOException An IO error occured
	 * @throws EOFException There is no successor-node
	 */
	public abstract String get_next() throws EOFException, IOException;

	/**
	 * Frees internal resources. Call once before the object loses its scope.
	 */
	public abstract void close();
}