/*
Krake Webcrawlerframework
Copyright (C) LS2 TU Dortmund (http://ls2-www.cs.tu-dortmund.de)

This file is part of the Krake Webcrawlerframework.
The sourcecode and data files provided come without any warranty. Use at your own risk!

The sourcecode is the intellectual property of its authors. You have permission to download
and use this sourcecode for the sake of inspection and for private and academic use. You are
not allowed to redistribute this sourcecode without our permission. If you want to use the
sourcecode in any other way as granted above,
please contact Christian Sohler <christian.sohler[at]tu-dortmund.de>.
*/
package krake.file;

import krake.Common;

import java.io.EOFException;
import java.io.IOException;
import java.io.InputStream;

/**
 * Basis for all classes that access the krake content file as a data stream. Stream access requires little to no extra memory
 * and is the fastest method if the entrire file should be read.
 * @author Thomas Nitschke
 */
public abstract class ContentStream extends Content
{
	protected abstract InputStream getIn();
	protected int m_cur_size;
	
	/**
	 * Reads the next content-string from the file.
	 * Return null on EOF.
	 * @return String or null
	 * @throws IOException
	 */
	public String next() throws IOException
	{
		String output = null;
		
		// we are at the eof mark
		if(m_cur_size == -1)
			return null;
		else if(m_cur_size != 0)// actual content
		{
			byte[] buffer = new byte[m_cur_size];
			readBuffer(getIn(), buffer);
			
			output = uncompress(buffer);
		}
		else // empty content
			output = "";
		
		int size = readSize(getIn());
		if(size != m_cur_size)
			throw new IOException("Corrupted file");

		// read the size of the next chunk
		try
		{
			m_cur_size = readSize(getIn());
		}
		catch(EOFException e)
		{
			// remember that we hit eof
			m_cur_size = -1;
		}
		return output;
	}

	/**
	 * Frees internal resources. Call once before the object loses its scope.
	 */
	public abstract void close();

	
	protected int readSize(InputStream in) throws IOException
	{
		byte[] integer = new byte[4];
		readBuffer(in, integer);
		return Common.intFromBytes(integer);
	}

	protected void readBuffer(InputStream stream, byte[] buffer) throws IOException
	{
		readBuffer(stream, buffer, buffer.length);
	}

	protected void readBuffer(InputStream stream, byte[] buffer, int to_read) throws IOException
	{
		int off = 0;
		int missing = to_read;
		int read;

		while(missing > 0)
		{
			read = stream.read(buffer, off, missing);
			if(read == -1)
				throw new EOFException();

			off += read;
			missing -= read;
		}
	}
}
