/*
Krake Webcrawlerframework
Copyright (C) LS2 TU Dortmund (http://ls2-www.cs.tu-dortmund.de)

This file is part of the Krake Webcrawlerframework.
The sourcecode and data files provided come without any warranty. Use at your own risk!

The sourcecode is the intellectual property of its authors. You have permission to download
and use this sourcecode for the sake of inspection and for private and academic use. You are
not allowed to redistribute this sourcecode without our permission. If you want to use the
sourcecode in any other way as granted above,
please contact Christian Sohler <christian.sohler[at]tu-dortmund.de>.
*/
package krake.file;

import krake.Common;
import krake.KrakeConfig;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FSDataInputStream;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.RawComparator;

import java.io.*;


/**
 * Bijective mapping between databases <code>byte[]</code> hash IDs and the
 * graphs integer node IDs. Both, the integer IDs as well as the hash IDs are
 * sorted. Thus int->hash is a simple lookup in the mapping array in time O(1)
 * and hash->int is performed by a binary search of the array content in time
 * O(log n).
 * 
 * @author Alexander Munteanu, Thomas Nitschke
 */
public class URLMapping
{
	private final Log LOG = LogFactory.getLog(URLMapping.class);
	private byte[] int2hash_map;

	/**
	 * Helperclass for key comparsions
	 */
	public static class ByteArrayComparator implements RawComparator<byte[]>
	{
		public ByteArrayComparator()
		{
			super();
		}

		public int compare(byte[] left, byte[] right)
		{
			return compareTo(left, right);
		}

		public int compare(byte[] b1, int s1, int l1, byte[] b2, int s2, int l2)
		{
			return compareTo(b1, s1, l1, b2, s2, l2);
		}

		public static int compareTo(final byte[] left, final byte[] right)
		{
			return compareTo(left, 0, left.length, right, 0, right.length);
		}

		public static int compareTo(byte[] b1, int s1, int l1, byte[] b2, int s2, int l2)
		{
			int end1 = s1 + l1;
			int end2 = s2 + l2;
			for(int i = s1, j = s2; i < end1 && j < end2; i++, j++)
			{
				int a = (b1[i] & 0xff);
				int b = (b2[j] & 0xff);
				if(a != b)
				{
					return a - b;
				}
			}
			return l1 - l2;
		}
	}

	/**
	 * Private CTor which enforces the useage of the static
	 * loading functions.
	 */
	private URLMapping()
	{
	}

	/**
	 * Creates an new mapping object with a given mapping.
	 * 
	 * @param map
	 *            mapping to be stored in the object
	 */
	public URLMapping(byte[] map)
	{
		int2hash_map = map;
	}

	/**
	 * Creates an new mapping object with a given mapping stored in a file.
	 * 
	 * @param filename Path to the local mapping file
	 */
	public static URLMapping openLocal(String filename) throws IOException
	{
		InputStream in = new BufferedInputStream(new FileInputStream(filename));
		URLMapping mapping = new URLMapping();
		mapping.loadMapping(in);
		in.close();
		return mapping;
	}

	/**
	 * Creates an new mapping object with a given mapping stored in a file.
	 *
	 * @param filename Path to the dfs mapping file
	 */
	public static URLMapping openDFS(FileSystem fs, Path dfspath) throws IOException
	{
		FSDataInputStream in = fs.open(dfspath);
		URLMapping mapping = new URLMapping();
		mapping.loadMapping(in);
		in.close();
		return mapping;
	}

	/**
	 * File loader for mapping files.
	 * 
	 * @param filename
	 *            filename of the mapping to be loaded
	 * @throws IOException
	 */
	private void loadMapping(InputStream in) throws IOException
	{
		LOG.info("Opener: Opening Mapping File");
		DataInputStream dis = new DataInputStream(in);

		int size = dis.readInt();
		LOG.info("Opener: Total Hash Entries: " + size);
		byte[] ret = new byte[size * 20];

		int total = 0;
		while(total < ret.length)
		{
			int cnt = dis.read(ret, total, ret.length - total);
			if(cnt < 0)
				throw new IOException("Opener: Unexpected EOF while reading file.");
			total += cnt;
		}

		LOG.info("Opener: Mapping File read done");
		int2hash_map = ret;
	}

	/**
	 * Gives access to the size of the mapping.
	 * 
	 * @return number of IDs stored in the mapping
	 */
	public int getSize()
	{
		return int2hash_map.length / 20;
	}

	/**
	 * Maps int->hash by a simple array lookup.
	 * 
	 * @param id
	 *            int ID
	 * @return hash ID
	 */
	public byte[] int2hash(int id)
	{
		byte[] hash = new byte[20];
		System.arraycopy(int2hash_map, id * 20, hash, 0, 20);
		return hash;

	}


	/**
	 * Maps hash->int by a binary search in the mapping.
	 * 
	 * @param hash
	 *            hash ID
	 * @return int ID
	 */
	public int hash2int(byte[] hash)
	{
		return binarySearch(int2hash_map, hash);
	}

	/**
	 * Performs a binary search for the given key (hash ID) in the mapping
	 * array.
	 * 
	 * @param arr
	 *            the mapping
	 * @param key
	 *            hash ID to search for
	 * @return int ID (rank of the hash ID in the array)
	 */
	private int binarySearch(byte[] arr, byte[] key)
	{
		RawComparator<byte[]> comparator = new ByteArrayComparator();
		int low = 0;
		int high = (arr.length - 1) / 20;

		while(low <= high)
		{
			int mid = (low + high) >>> 1;
			int cmp = comparator.compare(key, 0, 20, arr, mid * 20, 20);
			if(cmp > 0)
				low = mid + 1;
			else if(cmp < 0)
				high = mid - 1;
			else
				return mid;
		}
		return -(low + 1);
	}

	/**
	 * Simple selftest method.
	 */
	public void mappingTest()
	{
		LOG.info("start of mapping test");
		LOG.info("size: " + getSize());
		LOG.info("fetch data from mapping");
		for(int i = 0; i < getSize(); i++)
		{
			byte[] hash = int2hash(i);
			if(hash2int(hash) != i)
			{
				LOG.error("Error mapping entry: " + i + " / " + hash2int(hash));
			}
		}
		LOG.info("end of mapping test");
	}

	/**
	 * Quicktest main for debugging
	 */
	public static void main(String[] args) throws IOException
	{
		KrakeConfig.load();
		Configuration config = Common.getDefaultHadoopConfig();
		FileSystem fs = FileSystem.get(config);
		
		URLMapping map = URLMapping.openDFS(fs, new Path("totalexport/id_to_hash.fbin"));
		map.mappingTest();
	}
}
