/*
Krake Webcrawlerframework
Copyright (C) LS2 TU Dortmund (http://ls2-www.cs.tu-dortmund.de)

This file is part of the Krake Webcrawlerframework.
The sourcecode and data files provided come without any warranty. Use at your own risk!

The sourcecode is the intellectual property of its authors. You have permission to download
and use this sourcecode for the sake of inspection and for private and academic use. You are
not allowed to redistribute this sourcecode without our permission. If you want to use the
sourcecode in any other way as granted above,
please contact Christian Sohler <christian.sohler[at]tu-dortmund.de>.
*/
package krake.graph;

import krake.file.URLMapping;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.fs.FSDataInputStream;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;

import java.io.*;
import java.util.Arrays;
import java.util.Random;


/**
 * Adjacency list representation of the web graph.
 * 
 * @author Alexander Munteanu, Christopher Schroeder, Thomas Nitschke
 */
public class KrakeWebGraph implements IWebGraph
{
	private static final Log LOG = LogFactory.getLog(KrakeWebGraph.class);
	Random rand = new Random();
	private URLMapping urlmapping = null;
	private int[][] nodes = null;
	private long numberOfEdges = 0;

	/**
	 * Empty std. constructor.
	 */
	public KrakeWebGraph()
	{
	}
	
	protected KrakeWebGraph(DataInputStream in, long file_size) throws IOException
	{
		int[] tmp = new int[1];// call by reference parameter
		nodes = loadGraphNodes(in, file_size, tmp);
		numberOfEdges = tmp[0];
	}
	
	public static int[][] loadGraphNodes(DataInputStream in, long file_size, int[] out_numberOfEdges) throws IOException
	{
		int size = in.readInt(); // the saved number of nodes
		long fsize = file_size / 4; // the filesize in int

		LOG.info("WebGraph: " + size + " nodes in file");

		LOG.info("Creating array...");
		int[][] nodes = new int[size][];
		LOG.info("done");

		long bytecount = 1;

		LOG.info("Filling array");
		int no_in = 0;
		int no_out = 0;
		out_numberOfEdges[0] = 0;

		int cnt = 0;
		for(int i = 0; i < size; i++)
		{

			// display loading
			if(++cnt % (size / 10) == 0)
			{
				cnt = 0;
				LOG.info("...");
			}
			// eof check
			if(++bytecount > fsize)
			{
				LOG.error("Unexpected end of file, the data is incorrect");
				return nodes;
			}

			// read link count
			int onesize = in.readInt();
			if(onesize == 0)
				throw new IOException("Zero onesize, invalid data");
			
			out_numberOfEdges[0] += onesize - 1;
			
			int node[] = new int[onesize];
			
			// read the inlink-count
			node[0] = in.readInt();
			++bytecount;
			
			// read links
			for(int j = 1; j < onesize; j++)
			{
				// eof check
				if(++bytecount > fsize)
				{
					LOG.error("Unexpected end of file, the data is incorrect");
					return nodes;
				}

				node[j] = in.readInt();
			}
			
			int readable = onesize;
			for(int j = 1; j < node.length; ++j)
			{
				// invalid link
				if(node[j] >= size)
				{
					// check if this is an inlink, if so, decrement the inlink count
					if(j - 1 < node[0])
						--node[0];
					
					// mark as invalid
					node[j] = -1;
					
					--readable;
				}
			}
			
			nodes[i] = new int[readable];
			int wo = 0;
			for(int j = 0; j < onesize; ++j)
			{
				if(node[j] == -1)
					continue;
				else
				{
					nodes[i][wo] = node[j];
					++wo;
				}
			}
			if(wo != readable)
				System.out.println("Fix failed");
			
			if(nodes[i][0] == 0)
				++no_in;
			if(nodes[i].length - nodes[i][0] - 1 == 0)
				++no_out;
			
			node = null;
		}
		
		LOG.info(no_in + " node without any inlinks, " + no_out + " nodes without any outlinks");
		if(bytecount != fsize)
			LOG.warn("Unreaded remaining data at the end of the file");

		LOG.info("Webgraph loaded");
		return nodes;
	}
	
	public static KrakeWebGraph loadLocal(String path) throws IOException
	{
		DataInputStream in = null;
		try
		{
			File file = new File(path);
			in = new DataInputStream(new BufferedInputStream(new FileInputStream(path)));
			KrakeWebGraph graph = new KrakeWebGraph(in, file.length());
			return graph;
		}
		finally
		{
			if(in != null)
				in.close();
		}
	}
	
	public static KrakeWebGraph loadDFS(FileSystem fs, Path path) throws IOException
	{
		FSDataInputStream in = null;
		try
		{
			in = fs.open(path);
			long size = fs.getFileStatus(path).getLen();
			KrakeWebGraph graph = new KrakeWebGraph(in, size);
			return graph;
		}
		finally
		{
			if(in != null)
				in.close();
		}
	}

	/**
	 * Setter for setting the adjacency lists to a given content.
	 * 
	 * @param nodes
	 *            adjacency lists as two dimensional array
	 */
	public void setNodes(int[][] nodes)
	{
		this.nodes = nodes;
	}

	/**
	 * Sets the internal mapping to a given one.
	 * 
	 * @param map
	 *            the mapping
	 */
	public void setMapping(URLMapping map)
	{
		urlmapping = map;
	}

	/**
	 * Converts hash value to mapping id.
	 * 
	 * @param hash
	 *            the hash value
	 * @return the graph id from the mapping
	 */
	public int hash2int(byte[] hash)
	{
		return urlmapping.hash2int(hash);
	}

	/**
	 * Converts graph ID to hash id.
	 * 
	 * @param id
	 *            the integer ID
	 * @return the hash of the node
	 */
	public byte[] int2hash(int id)
	{
		return urlmapping.int2hash(id);
	}

	/**
	 * Getter for the whole adjacency list of a node referenced by its hash.
	 * 
	 * @param hash
	 *            hash of the node
	 * @return adjacency list of the given node
	 */
	public int[] getNode(byte[] hash)
	{
		int inode = urlmapping.hash2int(hash);
		if(inode < 0)
			return null;
		return getNode(inode);
	}

	/**
	 * Getter for the whole adjacency list of a given node.
	 * 
	 * @param id
	 *            node ID
	 * @return adjacency list of the given node
	 */
	public int[] getNode(int id)
	{
		return nodes[id];
	}

	/**
	 * Checks whether there exists an outgoing edge from node i to node j.
	 * 
	 * @param i
	 *            source node
	 * @param j
	 *            target node
	 * @return <b>true</b> iff there exists an edge from node i to node j
	 */
	@Override
	public boolean isEdge(int i, int j)
	{
		int[] nodei = nodes[i];
		return Arrays.binarySearch(nodei, nodei[0] + 1, nodei.length, j) >= 0;
	}

	/**
	 * Checks whether node i has an incoming edge from node j.
	 * 
	 * @param i
	 *            target node
	 * @param j
	 *            source node
	 * @return <b>true</b> iff there exists an edge from node j to node i
	 */
	public boolean isEdgeRev(int i, int j)
	{
		int[] nodei = nodes[i];
		return Arrays.binarySearch(nodei, 1, nodei[0] + 1, j) >= 0;
	}

	/**
	 * Checks whether there exists an outgoing edge from node i to node j where
	 * i and j are given by their hash values.
	 * 
	 * @param hash1
	 *            source node
	 * @param hash2
	 *            target node
	 * @return <b>true</b> iff thequeried edge exists
	 */
	public boolean isEdge(byte[] hash1, byte[] hash2)
	{
		int i = urlmapping.hash2int(hash1);
		int j = urlmapping.hash2int(hash2);
		if(i >= 0 && j >= 0)
			return isEdge(i, j);
		else
			return false;
	}

	/**
	 * Getter for the number of nodes of the graph.
	 * 
	 * @return number of nodes
	 * @see krake.graph.IWebGraph#getNumberOfNodes()
	 */
	public int getNumberOfNodes()
	{
		return nodes.length;
	}

	/**
	 * Getter for the number of edges contained in the graph.
	 * 
	 * @return number of edges of the graph
	 */
	@Override
	public long getNumberOfEdges()
	{
		return numberOfEdges;
	}

	/**
	 * Computes the averall degree (indegree + outdegree) of a given node.
	 * 
	 * @param inode
	 *            a node for which to compute the degree
	 * @return degree of the given node
	 */
	@Override
	public int getDegree(int inode)
	{
		return nodes[inode].length - 1;
	}

	/**
	 * Comutes the indegree of a given node.
	 * 
	 * @param inode
	 *            node ID for which to compute the indegree
	 * @return indegree of the given node
	 */
	@Override
	public int getIndegree(int inode)
	{
		return nodes[inode][0];
	}

	/**
	 * Comutes the outdegree of a given node.
	 * 
	 * @param inode
	 *            node ID for which to compute the outdegree
	 * @return outdegree of the given node
	 */
	@Override
	public int getOutdegree(int inode)
	{
		return nodes[inode].length - nodes[inode][0] - 1;
	}

	/**
	 * Query to the inlinks of a node.
	 * 
	 * @param inode
	 *            node for which to return the inlinks
	 * @return inlink adjacency list of the given node
	 * @see krake.graph.IWebGraph#getInlinks(int)
	 */
	public int[] getInlinks(int inode)
	{
		int[] inlinks = new int[getIndegree(inode)];
		System.arraycopy(nodes[inode], 1, inlinks, 0, inlinks.length);
		return inlinks;
	}

	/**
	 * Query to the outlinks of a node.
	 * 
	 * @param inode
	 *            node for which to return the outlinks
	 * @return outlink adjacency list of the given node
	 * @see krake.graph.IWebGraph#getOutlinks(int)
	 */
	public int[] getOutlinks(int inode)
	{
		int[] outlinks = new int[getOutdegree(inode)];
		System.arraycopy(nodes[inode], nodes[inode][0] + 1, outlinks, 0, outlinks.length);
		return outlinks;
	}

	/**
	 * Query to the whole adjacency list of a node.
	 * 
	 * @param inode
	 *            node for which to return all links
	 * @return whole adjacency list of the given node
	 * @see krake.graph.IWebGraph#getAlllinks(int)
	 */
	public int[] getAlllinks(int inode)
	{
		int[] alllinks = new int[getDegree(inode)];
		System.arraycopy(nodes[inode], 1, alllinks, 0, alllinks.length);
		return alllinks;
	}

	/**
	 * Computes the average degree of the graph (based on the outdegree of the
	 * nodes).
	 * 
	 * @return average degree
	 */
	public double averageDegree()
	{
		int cnt = 0;
		for(int i = 0; i < nodes.length; i++)
		{
			cnt += nodes[i].length - nodes[i][0];
		}
		return ((double) cnt / (double) nodes.length) - 1.0d;
	}

	/**
	 * Computes the average degree of the graph (based on the degree of the
	 * nodes).
	 * 
	 * @return average degree
	 */
	public double averageDegree2()
	{
		int cnt = 0;
		for(int i = 0; i < nodes.length; i++)
		{
			cnt += nodes[i].length;
		}
		return ((double) cnt / (double) nodes.length) - 1.0d;
	}

	/**
	 * Average (out)degree estimator. Outputs a value that by
	 * hoeffding.inequality approximates the average degree within an additive
	 * error of +- epsilon*n with probability at most delta.
	 * 
	 * @param epsinv
	 *            = 1/epsilon
	 * @param deltainv
	 *            = 1/delta
	 * @return average degree estimator
	 */
	public double averageDegreeEst(int epsinv, int deltainv)
	{
		long cnt = 0;
		int rounds = (int) ((epsinv * Math.log(2.0d * (double) deltainv) / 2.0d) + 1.0d);
		for(int i = 0; i < rounds; ++i)
		{
			int node = rand.nextInt(getNumberOfNodes());
			cnt += nodes[node].length - nodes[node][0];
		}
		return ((double) cnt / (double) rounds) - 1.0d;
	}

	/**
	 * Test method.
	 */
	public void correctCheck()
	{

		{
			LOG.info("WebGraph: Checking if all inlinks exists in outlinks too");
			int[] inlinks;
			for(int i = 0; i < nodes.length; i++)
			{
				inlinks = this.getInlinks(i);
				for(int j = 0; j < inlinks.length; j++)
				{
					if(!isEdge(inlinks[j], i))
					{ // Errorcheck
						LOG.info("WebGraph: Inlinks from " + i + " to " + inlinks[j] + " but no inlink from " + inlinks[j] + " to " + i);
						LOG.info("inlinks from " + i + ": ");
						for(int k : inlinks)
							LOG.info(k + ", ");


						int[] outlinks = getOutlinks(inlinks[j]);
						LOG.info("outlinks from " + inlinks[j] + ": ");
						for(int k : outlinks)
							LOG.info(k + ", ");
					}

				}
			}
			LOG.info("WebGraph: Check done");
		}

		{
			LOG.info("WebGraph: Checking if all outlinks exists in inlinks too");
			int[] outlinks;
			for(int i = 0; i < nodes.length; i++)
			{
				outlinks = this.getOutlinks(i);
				for(int j = 0; j < outlinks.length; j++)
				{
					if(!isEdgeRev(outlinks[j], i))
					{ // Errorcheck
						LOG.info("WebGraph: Outlink from " + i + " to " + outlinks[j] + " but no inlink from " + outlinks[j] + " to " + i);
						LOG.info("outlinks from " + i + ": ");
						for(int k : outlinks)
							LOG.info(k + ", ");
						

						int[] inlinks = getInlinks(outlinks[j]);
						LOG.info("inlinks from " + outlinks[j] + ": ");
						for(int k : inlinks)
							LOG.info(k + ", ");
						return;
					}

				}
			}
			LOG.info("WebGraph: Check done");
		}

		LOG.info("WebGraph: Checking something i don't remember :D");

		for(int i = 0; i < nodes.length; i++)
		{
			int[] inlinks = getInlinks(i);

			for(int j = 0; j < inlinks.length; j++)
			{
				if(getOutdegree(inlinks[j]) == 0)
				{
					arrayOut(getOutlinks(inlinks[j]));
				}
			}
		}

		LOG.info("Webgraph: Graph correct");
	}

	/**
	 * Output of a given integer array to stdout.
	 * 
	 * @param arr
	 *            an array
	 */
	private void arrayOut(int[] arr)
	{
		for(int i : arr)
		{
			System.out.print(i + ", ");
		}
		System.out.println();
	}

	public void saveLocal(String path) throws IOException
	{
		DataOutputStream out = new DataOutputStream(new FileOutputStream(path));
		
		out.writeInt(getNumberOfNodes());
		for(int[] row : nodes)
		{
			out.writeInt(row.length);
			for(int i : row)
				out.writeInt(i);
		}
		out.close();
	}

	/**
	 * Test method for average degree computation.
	 * 
	 * @param graph
	 *            graph for the computation
	 */
	public static void averageDegreeTest(KrakeWebGraph graph)
	{
		System.out.println("Webgraph: exact avg degree:");
		System.out.println(graph.averageDegree());
		int eps = 10000;
		int del = 60;
		int repetition = 10;
		System.out.println("Webgraph: avg estimator (" + repetition + " runs):");
		for(int i = 0; i < repetition; ++i)
			System.out.println(graph.averageDegreeEst(eps, del));
		System.out.println("Webgraph: max estimator (" + repetition + " runs):");
	}
}
