/*
Krake Webcrawlerframework
Copyright (C) LS2 TU Dortmund (http://ls2-www.cs.tu-dortmund.de)

This file is part of the Krake Webcrawlerframework.
The sourcecode and data files provided come without any warranty. Use at your own risk!

The sourcecode is the intellectual property of its authors. You have permission to download
and use this sourcecode for the sake of inspection and for private and academic use. You are
not allowed to redistribute this sourcecode without our permission. If you want to use the
sourcecode in any other way as granted above,
please contact Christian Sohler <christian.sohler[at]tu-dortmund.de>.
*/
package krake.graph;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import java.io.*;
import java.net.Socket;

/**
 * Client application for use with the krake web graph server.
 *
 * @author Christopher Schroeder
 */
public class WebGraphClient implements IWebGraph
{
	private final static Log LOG = LogFactory.getLog(WebGraphClient.class);

	private Socket sock;
	private DataInputStream in;
	private DataOutputStream out;
	private boolean active = false;


	/**
	 * Creates and initializes a new client and connects to the given server.
	 *
	 * @param server the server to connect to (e.g. "mud7")
	 * @throws IOException If the client cannot connect to the server
	 */
	public WebGraphClient(String server) throws IOException
	{
		sock = new Socket(server, WebGraphServer.PORT);
		sock.setSoTimeout(5000);
		in = new DataInputStream(new BufferedInputStream(sock.getInputStream()));
		out = new DataOutputStream(new BufferedOutputStream(sock.getOutputStream()));
		active = true;
	}

	/**
	 * Processes a query to the number of nodes of the underlying graph.
	 *
	 * @return the number of nodes
	 * @see krake.graph.IWebGraph#getNumberOfNodes()
	 */
	public int getNumberOfNodes()
	{
		try
		{
			out.writeInt(WebGraphServer.NUM_NODES);
			out.flush();
			return in.readInt();
		}
		catch(Exception e)
		{
			active = false;
			e.printStackTrace();
		}
		return -1;
	}

	/**
	 * Processes a query to the number of edges of the underlying graph.
	 *
	 * @return the number of edges
	 */
	public long getNumberOfEdges()
	{
		try
		{
			out.writeInt(WebGraphServer.NUM_EDGES);
			out.flush();
			return in.readLong();
		}
		catch(Exception e)
		{
			active = false;
			e.printStackTrace();
		}
		return -1;
	}

	@Override
	public int getDegree(int inode)
	{
		return 0;  //To change body of implemented methods use File | Settings | File Templates.
	}

	/**
	 * Processes a query to the indegree of a given node.
	 *
	 * @param inode node ID
	 * @return indegree of the node or -1 for unknown nodes
	 */
	public int getIndegree(int inode)
	{
		try
		{
			out.writeInt(WebGraphServer.INDEGREE_INODE);
			out.writeInt(inode);
			out.flush();
			return in.readInt();
		}
		catch(Exception e)
		{
			active = false;
			e.printStackTrace();
		}
		return -1;
	}

	/**
	 * Processes a query to the outdegree of a given node.
	 *
	 * @param inode node ID
	 * @return the outdegree of the node or -1 for unknown nodes
	 */
	public int getOutdegree(int inode)
	{
		try
		{
			out.writeInt(WebGraphServer.OUTDEGREE_INODE);
			out.writeInt(inode);
			out.flush();
			return in.readInt();
		}
		catch(Exception e)
		{
			active = false;
			e.printStackTrace();
		}
		return -1;
	}

	/**
	 * Processes a query to the inlink adjacency list of a given node.
	 *
	 * @param inode node ID
	 * @return inlink adjacency list of the node
	 * @see krake.graph.IWebGraph#getInlinks(int)
	 */
	public int[] getInlinks(int inode)
	{
		try
		{
			out.writeInt(WebGraphServer.INLINKS_INODE);
			out.writeInt(inode);
			out.flush();

			int[] ret = new int[in.readInt()];
			for(int i = 0; i < ret.length; i++)
			{
				ret[i] = in.readInt();
			}
			return ret;
		}
		catch(Exception e)
		{
			active = false;
			e.printStackTrace();
		}
		return null;
	}

	/**
	 * Processes a query to the outlink adjacency list of a given node.
	 *
	 * @param inode node ID
	 * @return outlink adjacency list of the node
	 * @see krake.graph.IWebGraph#getOutlinks(int)
	 */
	public int[] getOutlinks(int inode)
	{
		try
		{
			out.writeInt(WebGraphServer.OUTLINKS_INODE);
			out.writeInt(inode);
			out.flush();

			int[] ret = new int[in.readInt()];
			for(int i = 0; i < ret.length; i++)
			{
				ret[i] = in.readInt();
			}
			return ret;
		}
		catch(Exception e)
		{
			active = false;
			e.printStackTrace();
		}
		return null;
	}


	/**
	 * Processes a query to the whole adjacency list of a given node.
	 *
	 * @param inode node ID
	 * @return complete adjacency list (in- and outlinks) of the node
	 * @see krake.graph.IWebGraph#getAlllinks(int)
	 */
	public int[] getAlllinks(int inode)
	{
		try
		{
			out.writeInt(WebGraphServer.ALLLINKS_INODE);
			out.writeInt(inode);
			out.flush();

			int[] ret = new int[in.readInt()];
			for(int i = 0; i < ret.length; i++)
			{
				ret[i] = in.readInt();
			}
			return ret;
		}
		catch(Exception e)
		{
			active = false;
			e.printStackTrace();
		}
		return null;
	}

	/**
	 * Processes an adjacency query to find out whether there exists an edge from a given node to another.
	 *
	 * @param inode source node
	 * @param j	 target node
	 * @return <b>true</b> iff there exists an edge from node inode to node j
	 */
	public boolean isEdge(int inode, int j)
	{
		try
		{
			out.writeInt(WebGraphServer.IS_IEDGE);
			out.writeInt(inode);
			out.writeInt(j);
			out.flush();
			return in.readBoolean();
		}
		catch(Exception e)
		{
			active = false;
			e.printStackTrace();
		}
		return false;
	}

	/**
	 * Disconnects from the server.
	 */
	public void close()
	{
		if(!active)
			return;
		try
		{
			out.writeInt(WebGraphServer.CONN_ABORT);
			out.flush();
			sock.close();
		}
		catch(IOException ex)
		{
			LOG.error("Error during close", ex);
		}
	}
}
