/*
Krake Webcrawlerframework
Copyright (C) LS2 TU Dortmund (http://ls2-www.cs.tu-dortmund.de)

This file is part of the Krake Webcrawlerframework.
The sourcecode and data files provided come without any warranty. Use at your own risk!

The sourcecode is the intellectual property of its authors. You have permission to download
and use this sourcecode for the sake of inspection and for private and academic use. You are
not allowed to redistribute this sourcecode without our permission. If you want to use the
sourcecode in any other way as granted above,
please contact Christian Sohler <christian.sohler[at]tu-dortmund.de>.
*/
package krake.graph;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import java.io.*;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Server application for use with krake webgraph. Supports concurrently connected clients
 * and incoming queries.
 *
 * @author Alexander Munteanu
 */
public class WebGraphServer
{
	private static Log LOG = LogFactory.getLog(WebGraphServer.class);
	public static final int PORT = 23451;// 2326; // port the listens on

	private ServerSocket m_svsock;

	private IWebGraph webgraph;

	public static final int CONN_ABORT = -1;
	public static final int NUM_NODES = 0;
	public static final int NUM_EDGES = 1;
	public static final int INDEGREE_INODE = 2;
	public static final int OUTDEGREE_INODE = 3;
	public static final int INLINKS_INODE = 4;
	public static final int OUTLINKS_INODE = 5;
	public static final int IS_IEDGE = 6;
	public static final int ALLLINKS_INODE = 7;

	// GET_INODE, INLINK_INODE

	/**
	 * Thread for accepting new connections. Prevents timeouts on the client side.
	 *
	 * @author Alexander Munteanu
	 */
	private class AcceptThread extends Thread
	{
		@Override
		public void run()
		{
			while(true)
			{
				try
				{
					Socket client = m_svsock.accept();
					LOG.info("New client " + client.toString());
					RunThread runner = new RunThread(client);
					runner.start();
				}
				catch(IOException ex)
				{
					LOG.error("", ex);
				}
			}
		}
	}

	/**
	 * Main thread processes incoming queries.
	 *
	 * @author Alexander Munteanu
	 */
	private class RunThread extends Thread
	{
		Socket client;

		RunThread(Socket cl)
		{
			client = cl;
		}

		@Override
		public void run()
		{
			// open output sream
			DataInputStream in = null;
			DataOutputStream out = null;

			try
			{
				in = new DataInputStream(new BufferedInputStream(client
						.getInputStream()));
				out = new DataOutputStream(new BufferedOutputStream(client
						.getOutputStream()));
			}
			catch(IOException ex)
			{
				Logger.getLogger(WebGraphServer.class.getName()).log(
						Level.INFO, null, ex);
			}

			boolean leave = false;
			while(!leave && client.isConnected())
			{
				try
				{
					int inode;
					int[] links;

					int command = in.readInt();

					switch(command)
					{
						case CONN_ABORT:
							leave = true;
							break;
						case NUM_NODES:
							out.writeInt(webgraph.getNumberOfNodes());
							out.flush();
							break;
						case NUM_EDGES:
							out.writeLong(webgraph.getNumberOfEdges());
							out.flush();
							break;
						case INDEGREE_INODE:
							inode = in.readInt();
							out.writeInt(webgraph.getIndegree(inode));
							out.flush();
							break;
						case OUTDEGREE_INODE:
							inode = in.readInt();
							out.writeInt(webgraph.getOutdegree(inode));
							out.flush();
							break;
						case INLINKS_INODE:
							inode = in.readInt();
							links = webgraph.getInlinks(inode);
							out.writeInt(links.length);
							for(int link : links)
								out.writeInt(link);
							out.flush();
							links = null;
							break;
						case OUTLINKS_INODE:
							inode = in.readInt();
							links = webgraph.getOutlinks(inode);
							out.writeInt(links.length);
							for(int link : links)
								out.writeInt(link);
							out.flush();
							links = null;
							break;
						case IS_IEDGE:
							inode = in.readInt();
							int j = in.readInt();
							out.writeBoolean(webgraph.isEdge(inode, j));
							out.flush();
							break;
						case ALLLINKS_INODE:
							inode = in.readInt();
							links = webgraph.getAlllinks(inode);
							out.writeInt(links.length);
							for(int link : links)
								out.writeInt(link);
							out.flush();
							links = null;
							break;
						default:
							System.out
									.println("ERROR: unknown command: " + command);
					}
				}
				catch(IOException ex)
				{
					LOG.error("", ex);
					break;
				}
			}

			CloseClient(client);
			LOG.info("Client has disconnected " + client.toString());
		}// run
	}

	/**
	 * Terminates the socket connection to the given client.
	 *
	 * @param client the client to be disconnected
	 */
	private void CloseClient(Socket client)
	{
		try
		{
			client.close();
		}
		catch(IOException ex)
		{
			LOG.error("", ex);
		}
	}

	/**
	 * Loads web graph and mapping from files.
	 *
	 * @param graph Graph to use
	 * @throws IOException
	 */
	public WebGraphServer(IWebGraph graph) throws IOException
	{
		m_svsock = new ServerSocket(PORT);
		webgraph = graph;
	}

	/**
	 * Starts the server.
	 */
	public void run()
	{
		// start the accept-thread
		AcceptThread t = new AcceptThread();
		t.start();
	}

	/**
	 * Debug main
	 * Needs a path to a local webgraph as parameter and loads it into the webgraphserver.
	 */
	public static void main(String[] args)
	{
		if(args.length > 0)
		{
			Logger.getAnonymousLogger().setLevel(Level.ALL);
			WebGraphServer s;
			try
			{
				Logger.getAnonymousLogger().log(Level.INFO, "Loading webgraph");
				s = new WebGraphServer(KrakeWebGraph.loadLocal(args[0]));
				Runtime.getRuntime().gc();
				Runtime.getRuntime().gc();
				Runtime.getRuntime().gc();
				Runtime.getRuntime().gc();
				Logger.getAnonymousLogger().log(
						Level.INFO,
						"Free memory after loading webgraph: "
								+ Runtime.getRuntime().freeMemory() / 1024.0
								+ " KB");
				Logger.getAnonymousLogger().log(Level.INFO, "Starting server");
				Logger.getAnonymousLogger().log(Level.INFO,
						"Entering server-loop");
				s.run();
			}
			catch(IOException ex)
			{
				Logger.getAnonymousLogger().log(Level.SEVERE, null, ex);
			}
		} else
		{
			System.out.println("WebGraphServer: Please input filename of the graph/map as parameter");
		}
	}
}
