/*
Krake Webcrawlerframework
Copyright (C) LS2 TU Dortmund (http://ls2-www.cs.tu-dortmund.de)

This file is part of the Krake Webcrawlerframework.
The sourcecode and data files provided come without any warranty. Use at your own risk!

The sourcecode is the intellectual property of its authors. You have permission to download
and use this sourcecode for the sake of inspection and for private and academic use. You are
not allowed to redistribute this sourcecode without our permission. If you want to use the
sourcecode in any other way as granted above,
please contact Christian Sohler <christian.sohler[at]tu-dortmund.de>.
*/
package krake.keys;

import krake.KrakeConfig;

import java.io.IOException;
import java.sql.SQLException;

import org.apache.hadoop.conf.Configured;
import org.apache.hadoop.util.Tool;
import org.apache.hadoop.util.ToolRunner;

/**
 * Interface for keystorage-backends. Provides generic access to the keystorage.
 * @author Thomas Nitschke
 */
public abstract class KeyStorage
{
	/**
	 * Returns true if the key is known, i.e. the database has an entry for this key.
	 * @param key SHA1-Hash of the normalized URL
	 * @throws IOException
	 */
	public abstract boolean isKnown(String key) throws IOException;
	
	/**
	 * Returns the keyinfo for this key or null if the key is not known.
	 * @param key
	 */
	public abstract KeyInfo getKey(String key) throws IOException;
	
	/**
	 * Sets the keyinfo for the supplied key
	 * @param key SHA1-Hash of the normalized URL
	 * @param info Info to store for this key
	 */
	public abstract void setKey(String key, KeyInfo info) throws IOException;


    /**
     * Free any resources that are held by the KeyStorage. Should be the last call made for every instance.
     */
	public abstract void close();

    /**
     * Create a KeyStorage-instance as defined in the config (selection of backend and its parameters).
     * @return
     * @throws IOException
     */
	public static KeyStorage createFromConfig() throws IOException
	{
		if(!KrakeConfig.isLoaded())
			KrakeConfig.load();
		if(KrakeConfig.key_backend.equals("mysql"))
		{
			return new KeyStorageMySQL(KrakeConfig.servers, KrakeConfig.mysql_db, KrakeConfig.mysql_user, KrakeConfig.mysql_passwd);
		}
        else
            throw new IOException("No known backend for keystorage");
	}


	/**
	 * Does a one-time initialization for the current key-storagebackend.
	 * This call will _delete_ the current storage (if it exists) and create a blank storage.
	 * @throws IOException
	 */
	public static void initialSetup() throws IOException
	{
		KrakeConfig.load();

		if(KrakeConfig.key_backend.equals("mysql"))
		{
			try
			{
				KeyStorageMySQL ks = (KeyStorageMySQL)createFromConfig();
				ks.dropTables();
				ks.createTables();
			}
			catch(SQLException e)
			{
				e.printStackTrace();
				throw new IOException(e);
			}
		}
		else
			throw new IOException("No known backend for keystorage");
	}
	
	/**
	 * Helperclass for the initial table setup as hadoop job.
	 * @author Thomas Nitschke
	 */
	static class SetupTool extends Configured implements Tool
	{
		@Override
		public int run(String[] args) throws Exception
		{
			try
			{
				initialSetup();
				System.out.println("Initial setup successfull.");
			}
			catch(IOException e)
			{
				e.printStackTrace();
			}
			return 0;
		}
	}

	/**
	 * Start to setup the keystorage.
	 * @throws Exception 
	 */
	public static void main(String[] args) throws Exception
	{
		int ret = ToolRunner.run(new SetupTool(), args);
		System.exit(ret);
	}
}
