/*
Krake Webcrawlerframework
Copyright (C) LS2 TU Dortmund (http://ls2-www.cs.tu-dortmund.de)

This file is part of the Krake Webcrawlerframework.
The sourcecode and data files provided come without any warranty. Use at your own risk!

The sourcecode is the intellectual property of its authors. You have permission to download
and use this sourcecode for the sake of inspection and for private and academic use. You are
not allowed to redistribute this sourcecode without our permission. If you want to use the
sourcecode in any other way as granted above,
please contact Christian Sohler <christian.sohler[at]tu-dortmund.de>.
*/
package krake.keys;

import java.io.IOException;
import java.math.BigInteger;
import java.sql.*;

/**
 * MySQL-Backend for the keystorage. Agglomerates multiple MySQL-Servers into one distributed
 * keystorage. Provides basic query, write access as well as simple setup mechanisms.
 * @author Thomas Nitschke
 */
public class KeyStorageMySQL extends KeyStorage
{
	/// Max. number of keys that are managed by each server
	private BigInteger bnodes;

	/// Number of servers
	private int node_count;

	/// Connections to these servers
	private Connection[] connections;

	/**
	 * Creates a management-instance for the keystorage.
	 * @param servers List of server to connect to
	 * @param db Name of the database to use
	 * @param user Username
	 * @param passwd Password
	 * @throws IOException
	 */
	public KeyStorageMySQL(String[] servers, String db, String user, String passwd) throws IOException
	{
		node_count = servers.length;
		
		// calculcate the keyintervall that is handled by this instance directly
		BigInteger max = new BigInteger("FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF", 16);
		bnodes = max.divide(BigInteger.valueOf(node_count));
		
		// connect to all servers
		connections = new Connection[servers.length];
		int inst = 0;
		for(String server : servers)
		{
			try
			{
				connections[inst] = connect(server + "/" + db, user, passwd);
			}
			catch (SQLException e)
			{
				e.printStackTrace();
				throw new IOException(e);
			}
			++inst;
		}
	}
	
	private Connection connect(String server, String user, String passwd) throws SQLException
	{
		Connection con = DriverManager.getConnection("jdbc:mysql://" + server, user, passwd);
		con.setAutoCommit(true);
		return con;
	}
	
	private Connection getInstanceForKey(String skey)
	{
		BigInteger ikey = new BigInteger(skey, 16);
		int instance = ikey.divide(bnodes).intValue();
		if(instance >= node_count)
			throw new Error("Invalid instance id from key");
		return connections[instance];
	}

	/**
	 * Creates empty tables for the keystorage.
	 * @throws SQLException
	 */
	public void createTables() throws SQLException
	{
		for(Connection con : connections)
		{
			Statement st = con.createStatement();
			st.execute("CREATE TABLE krakekeys (hash CHAR(40) PRIMARY KEY, keyinfo TINYINT);");
		}
	}

	/**
	 * Drops the tables of the keystorage if the exist.
	 * Does not generate an error if the tables don't exist.
	 * @throws SQLException
	 */
	public void dropTables() throws SQLException
	{
		for(Connection con : connections)
		{
			Statement st = con.createStatement();
			st.execute("DROP TABLE IF EXISTS krakekeys");
		}
	}

	/**
	 * Empties the storage-tables but doesn't delete the tables themself.
	 * @see dropTables
	 * @throws SQLException
	 */
	public void deleteAllKeys() throws SQLException
	{
		for(Connection con : connections)
		{
			Statement st = con.createStatement();
			st.execute("DELETE FROM krakekeys");
		}
	}
	
	/**
	 * Performs a SQL-query for the specified key.
	 * @param con Connection to use
	 * @param key SHA1-Hexstring
	 * @return Index into KeyInfo Enum or -1
	 * @throws SQLException
	 */
	private int queryKey(Connection con, String key) throws SQLException
	{
		Statement st = con.createStatement();
		ResultSet rs = st.executeQuery("SELECT keyinfo from krakekeys where hash = \"" + key + "\";");
		if(rs.first())
			return rs.getInt(1);
		return -1;
	}

	/**
	 * Create or update the record for this key in the storage.
	 * @param con Connection to use
	 * @param key SHA1-Hexstring
	 * @param data KeyInfo encoded as integer
	 * @throws SQLException
	 */
	private void writeKey(Connection con, String key, int data) throws SQLException
	{
		Statement st = con.createStatement();
		st.execute("REPLACE into krakekeys VALUES (\"" + key + "\", " + data + ");");
	}
	
	@Override
	public boolean isKnown(String key) throws IOException
	{
		return getKey(key) != null;
	}

	@Override
	public void setKey(String key, KeyInfo info) throws IOException
	{
		try
		{
			writeKey(getInstanceForKey(key), key, info.ordinal());
		}
		catch (SQLException e)
		{
			e.printStackTrace();
			throw new IOException(e);
		}
	}

	@Override
	public KeyInfo getKey(String key) throws IOException
	{
		try
		{
			int ki = queryKey(getInstanceForKey(key), key);
			if(ki < 0)
				return null;
			return KeyInfo.values()[ki];
		}
		catch (SQLException e)
		{
			e.printStackTrace();
			throw new IOException(e);
		}
	}

	@Override
	public void close()
	{
		for(Connection con : connections)
		{
			if(con != null)
			{
				try
				{
					con.close();
				}
				catch (SQLException e)
				{
					e.printStackTrace();
				}
			}
		}
	}
}
