/*
Krake Webcrawlerframework
Copyright (C) LS2 TU Dortmund (http://ls2-www.cs.tu-dortmund.de)

This file is part of the Krake Webcrawlerframework.
The sourcecode and data files provided come without any warranty. Use at your own risk!

The sourcecode is the intellectual property of its authors. You have permission to download
and use this sourcecode for the sake of inspection and for private and academic use. You are
not allowed to redistribute this sourcecode without our permission. If you want to use the
sourcecode in any other way as granted above,
please contact Christian Sohler <christian.sohler[at]tu-dortmund.de>.
*/
package krake.mapreduce;

import krake.Common;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FSDataOutputStream;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapreduce.Reducer;

import java.io.IOException;
import java.util.Iterator;
import java.util.zip.Deflater;

/**
 *
 * @author Thomas Nitschke
 */
public class ExportReducer extends Reducer<Text, TextPairWritable, IntWritable, IntWritable>
{
	private final Log LOG = LogFactory.getLog(ExportReducer.class);

	public static final String PATH_OFFSETS = "krake.offsets";
	public static final String PATH_CONTENT = "krake.content";
	public static final String PATH_ID_TO_HASH = "krake.id_to_hash";
	public static final String PATH_HASH_TO_URL = "krake.hash_to_url";
	public static final String PAGE_COUNT = "krake.page_count";

	// io
	private FileSystem fs;
	private FSDataOutputStream id_to_hash_mapping;
	private FSDataOutputStream hash_to_url_mapping;
	private FSDataOutputStream content;
	private FSDataOutputStream offsets;
	private Deflater deflater = new Deflater();

	// id for the current page/node
	int id = 0;

	// first error that occured during setup
	private Exception setup_error = null;

	@Override
	protected void setup(Context ctx)
	{
		try
		{
			Configuration conf = ctx.getConfiguration();
			fs = FileSystem.get(conf);

			Path id2hash_path = new Path(conf.get(PATH_ID_TO_HASH));
			Path hash2url_path = new Path(conf.get(PATH_HASH_TO_URL));
			Path offsets_path = new Path(conf.get(PATH_OFFSETS));
			Path content_path = new Path(conf.get(PATH_CONTENT));
			if(fs.exists(id2hash_path) || fs.exists(hash2url_path) || fs.exists(offsets_path) || fs.exists(content_path))
				throw new IOException("Outputfiles already exist");

			hash_to_url_mapping = fs.create(hash2url_path, true);
			offsets = fs.create(offsets_path, true);
			content = fs.create(content_path, true);
			id_to_hash_mapping = fs.create(id2hash_path, true);

			// write pagecount in id <-> hash mapping
			int pages = ctx.getConfiguration().getInt(PAGE_COUNT, 0);
			if(pages == 0)
				throw new Exception("Pagecount invalid");

			id_to_hash_mapping.writeInt(pages);
		}
		catch(Exception e)
		{
			setup_error = e;
		}
	}

	@Override
	protected void cleanup(Context ctx)
	{
		try
		{
			content.close();
		}
		catch(IOException e)
		{
			e.printStackTrace();
		}
		try
		{
			offsets.close();
		}
		catch(IOException e)
		{
			e.printStackTrace();
		}
		try
		{
			hash_to_url_mapping.close();
		}
		catch(IOException e)
		{
			e.printStackTrace();
		}
		try
		{
			id_to_hash_mapping.close();
		}
		catch(IOException e)
		{
			e.printStackTrace();
		}
	}

	@Override
	protected void reduce(Text hash, Iterable<TextPairWritable> pages, Context ctx) throws IOException, InterruptedException
	{
		// test for successfull setup
		if(setup_error != null)
			throw new IOException(setup_error);


		Iterator<TextPairWritable> it = pages.iterator();
		if(it.hasNext())
		{
			TextPairWritable url_source = it.next();
			String url = url_source.m_first.toString();

			// id to hash mapping
			id_to_hash_mapping.write(Common.HextoByteArray(hash.toString()));

			// hash to url mapping

			hash_to_url_mapping.write(hash.toString().getBytes("UTF-8"));
			hash_to_url_mapping.write("\t".getBytes("UTF-8"));
			hash_to_url_mapping.write(url.replace("\n", "").getBytes("UTF-8"));// replace \n for safety
			hash_to_url_mapping.write("\n".getBytes("UTF-8"));

			// offset and content
			offsets.writeInt(id);
			offsets.writeLong(content.getPos());
			compressAndWriteContent(url_source.m_second.getBytes());

			++id;

			if(it.hasNext())
			{
				// this is very bad
				LOG.warn("Duplicated pages detected, truncate the urlmapping manually");
			}
		}
		else
			throw new IOException("Key without values in reducer");


	}

	private void compressAndWriteContent(byte[] raw_content) throws IOException
	{
		int size = raw_content.length;
		byte[] buffer = new byte[size*2];

		// compress content
		deflater.setInput(raw_content);
		deflater.finish();
		size = deflater.deflate(buffer);
		deflater.reset();

		// write compressed buffer with size-header/footer
		content.writeInt(size);
		content.write(buffer, 0, size);
		content.writeInt(size);
		buffer = null;
	}
}
