/*
Krake Webcrawlerframework
Copyright (C) LS2 TU Dortmund (http://ls2-www.cs.tu-dortmund.de)

This file is part of the Krake Webcrawlerframework.
The sourcecode and data files provided come without any warranty. Use at your own risk!

The sourcecode is the intellectual property of its authors. You have permission to download
and use this sourcecode for the sake of inspection and for private and academic use. You are
not allowed to redistribute this sourcecode without our permission. If you want to use the
sourcecode in any other way as granted above,
please contact Christian Sohler <christian.sohler[at]tu-dortmund.de>.
*/
package krake.mapreduce;

import krake.Common;
import krake.KrakeConfig;
import krake.Pair;
import krake.file.URLMapping;
import krake.keys.KeyInfo;
import krake.keys.KeyStorage;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapreduce.Counter;
import org.apache.hadoop.mapreduce.Mapper;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.HashSet;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class LinkerMapper extends Mapper<Text, Text, IntWritable, IntListWritable>
{
	Set<IntWritable> outlinks = new HashSet<IntWritable>(128);
	Set<Inlink> inlinks = new HashSet<Inlink>(128);

	Counter unmapped_url_count;

	URLMapping mapping = null;
	KeyStorage keys;

	Exception setup_error = null;
	
	final Log LOG = LogFactory.getLog(LinkerMapper.class);

	public void setup(Context ctx)
	{
		unmapped_url_count = ctx.getCounter(Common.CrawlCounters.UNMAPPED_URLS);

		try
		{
			keys = KeyStorage.createFromConfig();
		}
		catch(IOException e)
		{
			setup_error = e;
		}

		try
		{
			KrakeConfig.load();
			FileSystem fs = FileSystem.get(ctx.getConfiguration());
			mapping = URLMapping.openDFS(fs, new Path(KrakeConfig.id_to_hash_path));
		}
		catch(IOException e)
		{
			setup_error = e;
		}
	}
	
	public void map(Text surl, Text source, Context context) throws IOException, InterruptedException
	{
		// handle init errors
		if(setup_error != null)
			throw new IOException(setup_error);

		// prepare buffers
		inlinks.clear();
		outlinks.clear();
		
		String content = source.toString();
		String content_lc = content.toLowerCase();

		URL url = new URL(surl.toString());
		String hash = Common.normalize(url);
		int node = mapping.hash2int(Common.HextoByteArray(hash));
		if(node < 0)
			throw new IOException("URL not in URLMapping");

		mineURLsByPattern(url, content, content_lc, Common.LinkPattern, node, context);
		mineURLsByPattern(url, content, content_lc, Common.FramePattern, node, context);
		
		{	// write outlinks
			IntListWritable list = new IntListWritable(outlinks);
			context.write(new IntWritable(node + 1), list);
		}
		
		// write inlinks
		for(Inlink inlink : inlinks)
		{
			IntListWritable list = new IntListWritable();
			list.add(inlink.m_second);
			
			context.write(inlink.m_first, list);
		}
	}
	
	private void mineURLsByPattern(URL url, String content, String lowerCaseContent, Pattern p, int node, Context ctx) throws IOException
	{
		Matcher m = p.matcher(lowerCaseContent);
		int start = 0;
		String href;
		while(m.find(start))
		{
			href = null;
			start = m.end();

			// should never happen since the pattern matched
			if(m.groupCount() != 2)
			{
				System.out.println("Broken Pattern");
				continue;
			}

			// get the link from the original content
			// since the lowercase conversion destroys paths
			if(m.start(1) != -1 && m.end(1) != -1)
				href = content.substring(m.start(1), m.end(1)).trim();
			else if(m.start(2) != -1 && m.end(2) != -1)
				href = content.substring(m.start(2), m.end(2)).trim();

			// early skip for links that we don't want
			if(href.toLowerCase().startsWith("javascript") || href.startsWith("#"))
				continue;

			URL urlLink;
			try
			{
				urlLink = new URL(url, href);
				href = urlLink.toString();
				if(href.indexOf("#") != -1)
					href = href.substring(0, href.indexOf("#"));

				int quoteidx = href.indexOf("'");
				int dblquoteidx = href.indexOf("\"");
				quoteidx = (quoteidx == -1) ? href.length() : quoteidx;
				dblquoteidx = (dblquoteidx == -1) ? href.length() : dblquoteidx;

				if(quoteidx < dblquoteidx)
					href = href.substring(0, quoteidx);
				else
					href = href.substring(0, dblquoteidx);

				urlLink = new URL(url, href);
				href = urlLink.toString();
			}
			catch(MalformedURLException e)
			{
				continue;
			}

			// only look at http links that refer to other sites, not content
			if(!Common.isSupportedProtocol(urlLink.getProtocol()) || Common.isNonTextURL(urlLink))
				continue;

			// check against the crawlfilters
			for(Pattern pt : KrakeConfig.filters)
			{
				Matcher match = pt.matcher(href);
				if(match.find())
					continue;
			}

			// has the page been crawled
			String hash = Common.normalize(urlLink);
			if(keys.getKey(hash) != KeyInfo.CRAWLED)
				continue;
			
			int linked_node = mapping.hash2int(Common.HextoByteArray(hash));
			if(linked_node < 0)
			{
				unmapped_url_count.increment(1);
				//LOG.warn("Unmapped URL " + href);
			}
			else
			{
				outlinks.add(new IntWritable(linked_node+1));
				inlinks.add(new Inlink(new IntWritable(linked_node+1), new IntWritable(-(node+1))));
			}

			// offset by match
			start = m.end();
		}
	}

	/**
	 * Helperclass
	 */
	static class Inlink extends Pair<IntWritable, IntWritable> implements Comparable<Inlink>
	{
		public Inlink(IntWritable f, IntWritable s)
		{
			super(f, s);
		}

		@Override
		public int compareTo(Inlink o)
		{
			if(m_first.get() < o.m_first.get())
				return -1;
			if(m_first.get() > o.m_first.get())
				return 1;

			if(m_second.get() < o.m_second.get())
				return -1;
			if(m_second.get() > o.m_second.get())
				return 1;

			return 0;
		}
	}
}
