/*
Krake Webcrawlerframework
Copyright (C) LS2 TU Dortmund (http://ls2-www.cs.tu-dortmund.de)

This file is part of the Krake Webcrawlerframework.
The sourcecode and data files provided come without any warranty. Use at your own risk!

The sourcecode is the intellectual property of its authors. You have permission to download
and use this sourcecode for the sake of inspection and for private and academic use. You are
not allowed to redistribute this sourcecode without our permission. If you want to use the
sourcecode in any other way as granted above,
please contact Christian Sohler <christian.sohler[at]tu-dortmund.de>.
*/
package krake.mapreduce.formats;

import krake.file.ContentDFSRandomAccess;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapreduce.InputSplit;
import org.apache.hadoop.mapreduce.JobContext;
import org.apache.hadoop.mapreduce.TaskAttemptContext;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

/**
 * Hadoop inputformat for the krake content file. Set as input format for job that read a krake content file.
 * This format partitions the content file into inputsplits based on the number of available hadoop-nodes (5 hadoop-nodes -> 5 splits).
 * The behaviour of the format is controled by additional settings that have to be initialized before the job is started:
 * @see setContentFilePath
 * @see setOffsetFilePath
 * @see setFirstNode
 * @see setNodesToRead
 * @author Thomas Nitschke
 */
public class ContentFileInputFormat extends org.apache.hadoop.mapreduce.InputFormat<IntWritable, Text>
{
	final Log LOG = LogFactory.getLog(ContentFileInputFormat.class);
	public static final String CONTENT_FILE = "krake.contentfile";
	public static final String OFFSET_FILE = "krake.offsetfile";
	public static final String FIRST_INDEX = "krake.first_index";
	public static final String NODES_TO_READ = "krake.nodes_to_read";

	/**
	 * Sets the input krake content file for the configuration. The setting as effect only
	 * if its used to create the job!
	 * @param conf Configuration to change
	 * @param p Path
	 */
	public static void setContentFilePath(Configuration conf, Path p)
	{
		conf.set(CONTENT_FILE, p.toString());
	}

	/**
	 * Sets the input krake offset file for the configuration. The setting as effect only
	 * if its used to create the job!
	 * @param conf Configuration to change
	 * @param p Path
	 */
	public static void setOffsetFilePath(Configuration conf, Path p)
	{
		conf.set(OFFSET_FILE, p.toString());
	}

	/**
	 * Sets the first entry that will be read by the inputformat. The setting as effect only
	 * if its used to create the job!
	 * @param conf Configuration to change
	 * @param first Id/Index of the first entry, 0 is the first entry
	 */
	public static void setFirstNode(Configuration conf, int first)
	{
		conf.setInt(FIRST_INDEX, first);
	}

	/**
	 * Sets the number of entries that will be read by the inputformat. Set to a negative number if
	 * all entries in the content file should be read.
	 * @param conf Configuration to change
	 * @param nodes_to_read Number of nodes/entries that should be read or a negative number.
	 */
	public static void setNodesToRead(Configuration conf, int nodes_to_read)
	{
		conf.setInt(NODES_TO_READ, nodes_to_read);
	}

	protected class RecordReader extends org.apache.hadoop.mapreduce.RecordReader<IntWritable, Text>
	{
		private ContentDFSRandomAccess m_content = null;
		private int m_first;
		private Text m_value;
		private int m_read;
		private int m_count;

		public RecordReader()
		{
		}

		@Override
		public void initialize(InputSplit inputsplit, TaskAttemptContext context) throws IOException, InterruptedException
		{
			Configuration config = context.getConfiguration();
			m_content = new ContentDFSRandomAccess(config, new Path(config.get(OFFSET_FILE)), new Path(config.get(CONTENT_FILE)));
			
			ContentFileInputSplit split = (ContentFileInputSplit)inputsplit;
			
			
			m_count = split.getCount();
			m_first = split.getFirst();
			m_read = 0;
			m_value = null;
		}

		@Override
		public void close()
		{
			m_content.close();
		}

		@Override
		public float getProgress()
		{
			return (float)m_read / (float)m_count;
		}

		@Override
		public IntWritable getCurrentKey() throws IOException, InterruptedException
		{
			// -1 if nextKeyValue() wasn't called before
			return new IntWritable(m_first + m_read - 1);
		}

		@Override
		public Text getCurrentValue() throws IOException, InterruptedException
		{
			return m_value;
		}

		@Override
		public boolean nextKeyValue() throws IOException, InterruptedException
		{
			// no more nodes to read
			if(m_read == m_count)
				return false;
			
			// read the first node of the segment
			if(m_value == null && m_read == 0)
				m_value = new Text(m_content.get(m_first));
			else
				m_value = new Text(m_content.get_next());
			++m_read;
			return true;
		}

	}

	@Override
	public org.apache.hadoop.mapreduce.RecordReader<IntWritable, Text> createRecordReader(InputSplit split,
			TaskAttemptContext context)
	{
		return new RecordReader();
	}

	@Override
	public List<InputSplit> getSplits(JobContext context) throws IOException
	{
		ArrayList<InputSplit> splits = new ArrayList<InputSplit>();
		
		Configuration config = context.getConfiguration();
		Path contentfile = new Path(config.get(CONTENT_FILE));
		Path offsetfile = new Path(config.get(OFFSET_FILE));
		LOG.info("Calculating splits for " + contentfile + " and " + offsetfile);
		
		ContentDFSRandomAccess content = new ContentDFSRandomAccess(config, offsetfile, contentfile);
		
		int to_read = config.getInt(NODES_TO_READ, 0);
		int total = content.getEntryCount();
		
		// limit number of nodes to read if limit is set
		if(to_read > 0)
			total = to_read;
		
		// offset for splits
		int split_offset = config.getInt(FIRST_INDEX, 0);
		
		// TODO check for splits that go beyond the EOF
		
		int parallel_jobs = context.getConfiguration().getInt("mapred.jobtracker.taskScheduler.maxRunningTasksPerJob", 12);
		
		int split_size = total / parallel_jobs; //maybe 0????
		for(int i = 0; i < total/split_size; ++i)
			splits.add(new ContentFileInputSplit(split_size, split_offset + i*split_size));
		
		int rest = total % split_size;
		if(rest != 0)
			splits.add(new ContentFileInputSplit(rest, split_offset + total - rest));
		
		LOG.info("Parallel Jobs: " + parallel_jobs + " SplitSize: " + split_size + " Rest: " + rest);
		return splits;
	}
	
	public ContentFileInputFormat() {}

}
