/*
Krake Webcrawlerframework
Copyright (C) LS2 TU Dortmund (http://ls2-www.cs.tu-dortmund.de)

This file is part of the Krake Webcrawlerframework.
The sourcecode and data files provided come without any warranty. Use at your own risk!

The sourcecode is the intellectual property of its authors. You have permission to download
and use this sourcecode for the sake of inspection and for private and academic use. You are
not allowed to redistribute this sourcecode without our permission. If you want to use the
sourcecode in any other way as granted above,
please contact Christian Sohler <christian.sohler[at]tu-dortmund.de>.
*/
/**
 * Copyright (C) 2010 Doug Judd (Hypertable, Inc.)
 *
 * This file is part of Hypertable.
 *
 * Hypertable is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * Hypertable is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

package krake.mapreduce.formats;

import krake.KrakeConfig;
import krake.queue.Queue;
import krake.queue.QueueLocal;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapreduce.InputSplit;
import org.apache.hadoop.mapreduce.JobContext;
import org.apache.hadoop.mapreduce.TaskAttemptContext;

import java.io.BufferedReader;
import java.io.IOException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;


public class QueueLocalInputFormat extends org.apache.hadoop.mapreduce.InputFormat<Text, Text>
{
	final Log LOG = LogFactory.getLog(QueueLocalInputFormat.class);

	protected class RecordReader extends org.apache.hadoop.mapreduce.RecordReader<Text, Text>
	{
		private int m_lines_read;
		private int m_lines_count;

		private int[] m_queuefile_lines;
		private int m_current_queuefile = 0;
		private int m_queuefile_count;

		private Text m_key = null;
		private Text m_value = null;
		private BufferedReader m_reader = null;
		private QueueLocal m_queue = null;

		// true if we recurse into the nextKeyValue() function to recover from an error
		private boolean m_in_recovery_mode = false;

		/**
		 * Constructor
		 *
		 */
		public RecordReader()
		{
		}

		/**
		 * Initializes the reader.
		 *
		 * @param inputsplit The split to work with.
		 * @param context	The current task context.
		 * @throws java.io.IOException  When setting up the reader fails.
		 * @throws InterruptedException When the job is aborted.
		 * @see org.apache.hadoop.mapreduce.RecordReader#initialize(
		 *org.apache.hadoop.mapreduce.InputSplit,
		 *	  org.apache.hadoop.mapreduce.TaskAttemptContext)
		 */
		@Override
		public void initialize(InputSplit inputsplit, TaskAttemptContext context) throws IOException, InterruptedException
		{
			KrakeConfig.load();
			m_queue = (QueueLocal)Queue.createFromConfig();
			
			m_queuefile_lines = m_queue.getUnreadLines();
			m_queuefile_count = m_queue.getQueueFileCount();
			m_current_queuefile = 0; // queue always starts with file 0
			m_lines_count = KrakeConfig.batch_size; // amount of lines we want to read, there could be less

			if(m_queuefile_count == 0)
				throw new IOException("No queuefiles in split");
			
			// open first queuefile
			m_reader = m_queue.openQueueFile(m_current_queuefile);

			// read over the already processed lines (ie. the line offset)
			LOG.info("Seeking by reading to first line");
			int line_offset = m_queue.getLineOffset();
			for(int i = 0; i < line_offset; ++i)
			{
				if(m_reader.readLine() == null)
					throw new IOException("Unexpected EOF in queue file");
			}
		}

		/**
		 * Closes the split.
		 *
		 * @see org.apache.hadoop.mapreduce.RecordReader#close()
		 */
		@Override
		public void close()
		{
			try
			{
				if(m_reader != null)
					m_reader.close();
			}
			catch(Exception e)
			{
				e.printStackTrace();
				LOG.error("Closing the queue-reader failed", e);
			}
		}

		private void nextQueueFile() throws IOException
		{
			LOG.info("Switching from queuefile " + m_current_queuefile + " to " + (m_current_queuefile+1));
			++m_current_queuefile;

			// this should never happen
			assert(m_current_queuefile >= m_queuefile_count);

			// close old file and open the new one
			try
			{
				m_reader.close();
			}
			catch(IOException e)
			{
				e.printStackTrace();
				LOG.error("Closing the queue-reader failed", e);
			}

			m_reader = m_queue.openQueueFile(m_current_queuefile);
		}

		/**
		 * The current progress of the record reader through its data.
		 *
		 * @return A number between 0.0 and 1.0, the fraction of the data read.
		 * @see org.apache.hadoop.mapreduce.RecordReader#getProgress()
		 */
		@Override
		public float getProgress()
		{
			return (float) m_lines_read / (float) m_lines_count;
		}

		/**
		 * Returns the current key.
		 *
		 * @return The current key.
		 * @throws java.io.IOException
		 * @throws InterruptedException When the job is aborted.
		 * @see org.apache.hadoop.mapreduce.RecordReader#getCurrentKey()
		 */
		@Override
		public Text getCurrentKey() throws IOException, InterruptedException
		{
			return m_key;
		}

		/**
		 * Returns the current value.
		 *
		 * @return The current value.
		 * @throws java.io.IOException  When the value is faulty.
		 * @throws InterruptedException When the job is aborted.
		 * @see org.apache.hadoop.mapreduce.RecordReader#getCurrentValue()
		 */
		@Override
		public Text getCurrentValue() throws IOException, InterruptedException
		{
			return m_value;
		}

		/**
		 * Positions the record reader to the next record.
		 *
		 * @return <code>true</code> if there was another record.
		 * @throws java.io.IOException  When reading the record failed.
		 * @throws InterruptedException When the job was aborted.
		 * @see org.apache.hadoop.mapreduce.RecordReader#nextKeyValue()
		 */
		@Override
		public boolean nextKeyValue() throws IOException, InterruptedException
		{
			if(m_reader == null)
				throw new IOException("Reader is null");

			// we are done (all lines are read or queue is empty)
			if(m_lines_read == m_lines_count || m_current_queuefile == m_queuefile_count)
			{
				LOG.debug("Reached end of queue");
				return false;
			}

			String line = m_reader.readLine();
			if(line == null)
				throw new IOException("Unexpected EOF in queue file");
			
			++m_lines_read;
			--m_queuefile_lines[m_current_queuefile];

			// we need to switch to the next queuefile
			if(m_queuefile_lines[m_current_queuefile] == 0)
			{
				// sanity-check: the next read should return null
				String test = m_reader.readLine();
				if(test != null)
				{
					LOG.warn("Queuefile " + m_current_queuefile + " is processed by not empty");
					// treat the file as empty anyways, but something is wrong!
				}

				// switch to the next file, if we have one
				if((m_current_queuefile + 1) < m_queuefile_count)
					nextQueueFile();
				else
				{
					// current k/v pair is the last in the queue
					m_current_queuefile = m_queuefile_count;
				}
			}

			// process line, handle invalid URLs
			try
			{
				line = line.trim(); // just to be safe
				URL u = new URL(line);

				m_value = new Text(line);
				m_key = new Text(krake.Common.normalize(u));
			}
			catch(Exception e) // usually MalformedURLException
			{
				e.printStackTrace();
				LOG.error("Malformed URL in queue: '" + line + "'", e);
				
				if(m_in_recovery_mode)
				{
					// fix your input, sir!
					throw new IOException("Two malformed URLs in a row, giving up");
				}
				else
				{
					// try to read the next kv-pair
					m_in_recovery_mode = true; // enable recursion protection
					boolean next_result = nextKeyValue();
					m_in_recovery_mode = false;
					return next_result;
				}
			}
			return true;
		}
	}


	@Override
	public org.apache.hadoop.mapreduce.RecordReader<Text, Text> createRecordReader(
			InputSplit split, TaskAttemptContext context)
			throws IOException
	{
		return new RecordReader();
	}

	@Override
	public List<InputSplit> getSplits(JobContext context) throws IOException
	{
		// create a dummy split for each node -> one mapreduce process for each node
		List<InputSplit> splits = new ArrayList<InputSplit>(KrakeConfig.servers.length);
		for(int i = 0; i < KrakeConfig.servers.length; ++i)
			splits.add(new QueueLocalInputSplit(KrakeConfig.servers[i]));
		return splits;
	}

}
