/*
Krake Webcrawlerframework
Copyright (C) LS2 TU Dortmund (http://ls2-www.cs.tu-dortmund.de)

This file is part of the Krake Webcrawlerframework.
The sourcecode and data files provided come without any warranty. Use at your own risk!

The sourcecode is the intellectual property of its authors. You have permission to download
and use this sourcecode for the sake of inspection and for private and academic use. You are
not allowed to redistribute this sourcecode without our permission. If you want to use the
sourcecode in any other way as granted above,
please contact Christian Sohler <christian.sohler[at]tu-dortmund.de>.
*/
package krake.queue;

import krake.KrakeConfig;
import krake.mapreduce.formats.QueueLocalInputFormat;
import org.apache.hadoop.mapreduce.Job;

import java.io.IOException;

/**
 * Basic interface that provides access to the crawling queue. The queue is used as input for a crawl-job
 * and at the same time fed with new URLs by the crawler. Queues are handled locally! A crawl-mapper of a crawl-job
 * will read and write solely to the queue of the node it runs on.
 * Necessary management-tasks (@see preCrawl and @see postCrawl) have to be performed on each node.
 * @author Thomas Nitschke
 */
public abstract class Queue
{
	/**
	 * Prepares the local queue for a new crawl.
	 * Call once in each mapper-instance during setup.
	 * @see postCrawl
	 * @throws IOException
	 */
	public abstract void preCrawl() throws IOException;

	/**
	 * Add an URL to the queue.
	 * @param url
	 * @throws IOException
	 */
	public abstract void add(String url) throws IOException;

	/**
	 * Updates and cleans the queue-storage.
	 * Call once in each mapper-instance during shutdown.
	 * @see preCrawl
	 * @param processed_lines Number of lines/URL that were read during the crawl
	 * @throws IOException
	 */
	public abstract void postCrawl(int processed_lines) throws IOException;

	/**
	 * Create a queue-handle with the configuration parameters.
	 * @return
	 * @throws IOException
	 */
	public static Queue createFromConfig() throws IOException
	{
		if(KrakeConfig.queue_backend.equals("local"))
			return new QueueLocal();
		else
			throw new IOException("Unknown queue backend");
	}

	/**
	 * Sets the queue as input for the supplied hadoop job.
	 * @param job New Hadoop-Crawler
	 */
	public static void setupJob(Job job) throws IOException
	{
		if(KrakeConfig.queue_backend.equals("local"))
			job.setInputFormatClass(QueueLocalInputFormat.class);
		else
			throw new IOException("Unknown queue backend");
	}
}
