/*
Krake Webcrawlerframework
Copyright (C) LS2 TU Dortmund (http://ls2-www.cs.tu-dortmund.de)

This file is part of the Krake Webcrawlerframework.
The sourcecode and data files provided come without any warranty. Use at your own risk!

The sourcecode is the intellectual property of its authors. You have permission to download
and use this sourcecode for the sake of inspection and for private and academic use. You are
not allowed to redistribute this sourcecode without our permission. If you want to use the
sourcecode in any other way as granted above,
please contact Christian Sohler <christian.sohler[at]tu-dortmund.de>.
*/
package krake.queue;

import krake.KrakeConfig;

import java.io.*;
import java.util.ArrayList;

/**
 * Implementation of a queue that uses local files for storage.
 * @author Thomas Nitschke
 */
public class QueueLocal extends Queue
{
	int[] unread_lines; // amount of unread lines in the current list of queuefiles
	int line_offset; // amount of lines in the first queue file that were already read by previous runs
	BufferedWriter latest_queuefile = null; // file for new urls
	int latest_queuefile_lines = 0; // number of written lines in the new queuefile

	public QueueLocal() throws IOException
	{
		loadState();
	}

	public BufferedReader openQueueFile(int idx) throws IOException
	{
		assert(idx >= 0);
		if(idx >= unread_lines.length)
			throw new IOException("Queuefile index is out of bounds");

		return new BufferedReader(new FileReader(mkPath(Integer.toString(idx))));
	}

	public int getQueueFileCount()
	{
		return unread_lines.length;
	}

	/**
	 *
	 * @throws IOException
	 */
	@Override
	public void preCrawl() throws IOException
	{
		// open a new queue file in append-mode
		// we need append since a sequence of mappers might run on this node in one crawliteration
		latest_queuefile = new BufferedWriter(new FileWriter(mkPath("new"), true));
		latest_queuefile_lines = 0;
	}

	public final int[] getUnreadLines()
	{
		return unread_lines;
	}

	public final int getLineOffset()
	{
		return line_offset;
	}

	/**
	 * Build path to the queuefile of name filename
	 * @param filename
	 * @return
	 */
	public static String mkPath(String filename)
	{
		// TODO construct properly
		return KrakeConfig.queue_local_dir + "/" + filename;
	}

	@Override
	public void postCrawl(int processed_lines) throws IOException
	{
		// close the file since we want to move it later
		try
		{
			latest_queuefile.close();
		}
		catch(IOException e)
		{
			e.printStackTrace();
		}
		latest_queuefile = null;

		// check which files still contain lines that weren't read yet
		int obsolete_files = 0;
		line_offset = 0; // reset the lineoffset since the code might never reach the assigment in the else-part
		for(int i = 0; i < unread_lines.length; ++i)
		{
			if(processed_lines >= unread_lines[i])
			{
				++obsolete_files;
				processed_lines -= unread_lines[i];
			}
			else
			{
				line_offset = processed_lines;
				unread_lines[i] -= processed_lines;
				break;
			}
		}
		
		// delete obsolete queuefiles
		for(int i = 0; i < obsolete_files; ++i)
		{
			// construct path
			String path = mkPath(Integer.toString(i));
			
			// delete
			File f = new File(path);
			if(f.exists())
			{
				if(!f.delete())
					throw new IOException("Unable to delete queue files");
			}
		}

		// rearange remaining queuefiles
		// move i-th remaining file to (i - obsolete) file
		for(int i = obsolete_files; i < unread_lines.length; ++i)
		{
			// move
			String path_from = mkPath(Integer.toString(i));
			String path_to = mkPath(Integer.toString(i - obsolete_files));
			
			// move
			File f = new File(path_from);
			if(!f.renameTo(new File(path_to)))
				throw new IOException("Unable to reorganize queue file " + path_from);
		}

		// add the new queuefile
		if(latest_queuefile_lines > 0)
		{
			String path_from = mkPath("new");
			String path_to = mkPath(Integer.toString(unread_lines.length - obsolete_files));
			
			// move
			File f = new File(path_from);
			if(!f.renameTo(new File(path_to)))
				throw new IOException("Unable to reorganize queue file " + path_from);
		}
		
		// update lines array
		// copy remaining
		int[] new_unread = new int[unread_lines.length - obsolete_files + (latest_queuefile_lines > 0 ? 1 : 0)];
		System.arraycopy(unread_lines, obsolete_files, new_unread, 0, unread_lines.length - obsolete_files);
		
		// add new
		if(latest_queuefile_lines > 0)
			new_unread[new_unread.length - 1] = latest_queuefile_lines;
		unread_lines = new_unread;
		
		// save lines
		saveState();

		// reset, just to be safe
		latest_queuefile_lines = 0;
	}

	private void loadState() throws IOException
	{
		ArrayList<Integer> lines = new ArrayList<Integer>();
		
		BufferedReader r = new BufferedReader(new FileReader(mkPath("state")));
		while(true)
		{
			String line = r.readLine();
			if(line == null)
				break;
			else if(line.equals(""))
				continue; // skip empty lines
			try
			{
				lines.add(Integer.parseInt(line));
			}
			catch(NumberFormatException e)
			{
				throw new IOException(e); // translate exception
			}
		}
		r.close();

		// must be at least 2 lines, (lineoffset and lines of first queuefile)
		if(lines.size() < 1)
			throw new IOException("Queuestate-file is invalid (no entries found)");

		// first entry is the line_offset
		line_offset = lines.get(0);
		if(line_offset < 0)
			throw new IOException("Invalid lineoffset in queuestate-file");
		
		// lines
		unread_lines = new int[lines.size()-1];
		for(int i = 1; i < lines.size(); ++i)
		{
			unread_lines[i-1] = lines.get(i);
			if(unread_lines[i-1] <= 0)
				throw new IOException("Invalid line in queuestate. Zero or less unread lines in queuefile detected.");
		}
	}
	
	private void saveState() throws IOException
	{
		BufferedWriter wr = new BufferedWriter(new FileWriter(mkPath("state")));
		wr.write(Integer.toString(line_offset));
		wr.write("\n");
		for(int l : unread_lines)
		{
			wr.write(Integer.toString(l));
			wr.write("\n");
		}
		wr.close();
	}

	@Override
	public void add(String url) throws IOException
	{
		latest_queuefile.write(url);
		latest_queuefile.write("\n");
		++latest_queuefile_lines;
	}
}
