/*
Krake Webcrawlerframework
Copyright (C) LS2 TU Dortmund (http://ls2-www.cs.tu-dortmund.de)

This file is part of the Krake Webcrawlerframework.
The sourcecode and data files provided come without any warranty. Use at your own risk!

The sourcecode is the intellectual property of its authors. You have permission to download
and use this sourcecode for the sake of inspection and for private and academic use. You are
not allowed to redistribute this sourcecode without our permission. If you want to use the
sourcecode in any other way as granted above,
please contact Christian Sohler <christian.sohler[at]tu-dortmund.de>.
*/
package krake.search;

import org.apache.hadoop.fs.FSDataOutputStream;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.NullWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapreduce.RecordWriter;
import org.apache.hadoop.mapreduce.TaskAttemptContext;
import org.apache.hadoop.mapreduce.lib.output.TextOutputFormat;

import java.io.DataOutputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;

/**
 * Special Outputformat for the generation of inverted indices.
 * The ouputformat puts all values with the same key in one line, each value separated by the
 * normal keyValueSeparator-character (usually tab). This behaviour differs from other outputformats which
 * produce one line for each key-value pair. With a normal outputformat it would be necessary to collect all values for 
 * a key in the reducer and therefore most likely break the local memory limit.
 * @author Thomas Nitschke
 *
 */
public class DocListOutputFormat<K, V> extends TextOutputFormat<K, V>
{
	protected static class LineRecordWriterEx<K, V> extends LineRecordWriter<K, V>
	{
		private static final String utf8 = "UTF-8";
		private final byte[] keyValueSeparator;
		private static String last_key = null;
		private static final byte[] newline;
		
		static
		{
			try
			{
				newline = "\n".getBytes(utf8);
			}
			catch(UnsupportedEncodingException uee)
			{
				throw new IllegalArgumentException("can't find " + utf8 + " encoding");
			}
		}

		public LineRecordWriterEx(DataOutputStream out, String _keyValueSeparator)
		{
			super(out, _keyValueSeparator);
			try
			{
				keyValueSeparator = _keyValueSeparator.getBytes(utf8);
			}
			catch(UnsupportedEncodingException uee)
			{
				throw new IllegalArgumentException("can't find " + utf8 + " encoding");
			}
		}

		public LineRecordWriterEx(DataOutputStream out)
		{
			this(out, "\t");
		}

		public synchronized void write(K key, V value) throws IOException
		{
			boolean nullKey = key == null || key instanceof NullWritable;
			boolean nullValue = value == null || value instanceof NullWritable;
			if(nullKey || nullValue)
				return;
			
			if(!(key instanceof Text))
				return;
			
			if(last_key != null)
			{
				String skey = ((Text)key).toString();
				
				if(last_key.equals(skey))
				{
					out.write(keyValueSeparator);
					writeObject(value);
				}
				else
				{
					out.write(newline);
					writeObject(key);
					out.write(keyValueSeparator);
					writeObject(value);
					
					last_key = skey.toString();
				}
			}
			else
			{
				writeObject(key);
				out.write(keyValueSeparator);
				writeObject(value);
				
				last_key = ((Text)key).toString();
			}
		}

		private void writeObject(Object o) throws IOException
		{
			out.write(o.toString().getBytes(utf8));
		}
	}// LineRecordWriter

	public RecordWriter<K, V> getRecordWriter(TaskAttemptContext job)
	    throws IOException
	{
		boolean isCompressed = getCompressOutput(job);
		String keyValueSeparator = job.getConfiguration().get("mapred.textoutputformat.separator", "\t");
		if(!isCompressed)
		{
			Path file = getDefaultWorkFile(job, "");
			FileSystem fs = file.getFileSystem(job.getConfiguration());
			FSDataOutputStream fileOut = fs.create(file, false); // FIXME progress parameter?
			return new LineRecordWriterEx<K, V>(fileOut, keyValueSeparator);
		}
		else
			throw new UnsupportedOperationException("Compressed output is not supported by this outputformat");
	}
}
