/*
Krake Webcrawlerframework
Copyright (C) LS2 TU Dortmund (http://ls2-www.cs.tu-dortmund.de)

This file is part of the Krake Webcrawlerframework.
The sourcecode and data files provided come without any warranty. Use at your own risk!

The sourcecode is the intellectual property of its authors. You have permission to download
and use this sourcecode for the sake of inspection and for private and academic use. You are
not allowed to redistribute this sourcecode without our permission. If you want to use the
sourcecode in any other way as granted above,
please contact Christian Sohler <christian.sohler[at]tu-dortmund.de>.
*/
package krake.search;

import krake.mapreduce.formats.ContentFileInputFormat;
import org.apache.commons.cli.BasicParser;
import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Options;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.conf.Configured;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapreduce.Job;
import org.apache.hadoop.util.Tool;
import org.apache.hadoop.util.ToolRunner;

import java.io.IOException;

/**
 * Hadoop Crawler which generates an inverted index for a give KrakeContentFile.
 * See commandline help for usage info.
 * @author Thomas Nitschke
 *
 */
public class InvertedIndexJob extends Configured implements Tool
{
	public int run(String[] args) throws Exception
	{
		// parse cmdline
		Options opt = new Options();
		opt.addOption("content", true, "DFS path to the content file");
		opt.addOption("offset", true, "DFS path to the offset file");
		opt.addOption("output", true, "DFS path to the output directory"); 
		opt.addOption("help", false, "Print help");
		
		BasicParser parser = new BasicParser();
		CommandLine cl = parser.parse(opt, args);
		
		// print help if requested or arguments are missing
		if(cl.hasOption("help") || !(cl.hasOption("content") && cl.hasOption("offset") && cl.hasOption("output")))
		{
			HelpFormatter f = new HelpFormatter();
		    f.printHelp("InvertedIndex", opt);
		    return 0;
		}
		
		Job job = new Job(getConf(), "InvertedIndex");
		job.setJarByClass(InvertedIndexJob.class);

		// check inputfiles
		Path contentfile = new Path(cl.getOptionValue("content"));
		Path offsetfile = new Path(cl.getOptionValue("offset"));
		
		FileSystem fs = job.getWorkingDirectory().getFileSystem(job.getConfiguration());
		if(!(fs.exists(contentfile) && fs.exists(offsetfile)))
			throw new IOException("Inputfiles were not found");

		// setup job
		job.setMapperClass(InvertedIndexMapper.class);
		job.setReducerClass(InvertedIndexReducer.class);
		job.getConfiguration().set("mapred.child.java.opts", "-Xmx6500m"); // allow crawlmappers to use more memory
		
		job.setInputFormatClass(ContentFileInputFormat.class);
		job.getConfiguration().set(ContentFileInputFormat.CONTENT_FILE, contentfile.toString());
		job.getConfiguration().set(ContentFileInputFormat.OFFSET_FILE, offsetfile.toString());
		
		job.setOutputFormatClass(DocListOutputFormat.class);
		DocListOutputFormat.setOutputPath(job, new Path(cl.getOptionValue("output")));
		
		job.setMapOutputKeyClass(Text.class);
		job.setMapOutputValueClass(IntWritable.class);
		job.setOutputKeyClass(Text.class);
		job.setOutputValueClass(IntWritable.class);
		
		// just one reduce since we only want one output file
		job.setNumReduceTasks(1);

		job.submit();
		job.waitForCompletion(true);
		if(!job.isSuccessful())
			throw new Exception("InvertedIndex failed");
		return 0;
	}

	public static void main(String[] args) throws Exception
	{
		int res = ToolRunner.run(new Configuration(), new InvertedIndexJob(), args);
		System.exit(res);
	}
}
