/*
Krake Webcrawlerframework
Copyright (C) LS2 TU Dortmund (http://ls2-www.cs.tu-dortmund.de)

This file is part of the Krake Webcrawlerframework.
The sourcecode and data files provided come without any warranty. Use at your own risk!

The sourcecode is the intellectual property of its authors. You have permission to download
and use this sourcecode for the sake of inspection and for private and academic use. You are
not allowed to redistribute this sourcecode without our permission. If you want to use the
sourcecode in any other way as granted above,
please contact Christian Sohler <christian.sohler[at]tu-dortmund.de>.
*/
package krake.search;

import krake.Pair;
import krake.file.unused.IdToURLMapping;
import krake.graph.KrakeWebGraph;

import java.io.*;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;

/**
 * Management for various ranks.
 */
public class Rank
{
	protected itfPr to;
	protected double[] PRold;
	protected double[] PRnew;
	protected double[] Ranks;
	protected SimpleDateFormat dateformat = new SimpleDateFormat("HH:mm:ss ");;

	/**
	 * Loads pagerank values from a specified file.
	 * 
	 * @param in
	 *            stream from which the pagerank will be read
	 * @throws IOException
	 */
	public void loadPageRank(DataInputStream in) throws IOException
	{
		System.out.println("PageRank: Loading PR From file");
		Ranks = loadRankArray(in);
		System.out.println("PageRank: done");
	}
	
	public static double[] loadRankArray(DataInputStream in) throws IOException
	{
		int count = in.readInt();
		double[] ranks = new double[count];
		
		for(int i = 0; i < count; ++i)
			ranks[i] = in.readDouble();
		
		return ranks;
	}

	public void saveToFile(String filename) throws IOException
	{
		System.out.println("PageRank: Saving ranks to file");
		
		DataOutputStream out = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(filename)));
		out.writeInt(Ranks.length);
		for(double d : Ranks)
		{
			out.writeDouble(d);
		}
		out.close();
		
		System.out.println("PageRank: done");
	}

	/**
	 * Saves pagerank values to the specified file.
	 * 
	 * @param filename
	 *            file to be saved to
	 * @throws IOException
	 */
	public void saveToFile_Sorted(String filename) throws IOException
	{
		System.out.println("PageRank: Saving sorted ranks to file");
		
		ArrayList<Pair<Integer, Double> > ranks = new ArrayList<Pair<Integer, Double> >(Ranks.length);
		for(int i = 0; i < Ranks.length; ++i)
			ranks.add(new Pair<Integer, Double>(i, Ranks[i]));
		
		Collections.sort(ranks, new Comparator<Pair<Integer, Double>>() {
				@Override
				public int compare(Pair<Integer, Double> o1, Pair<Integer, Double> o2)
				{
					return Double.compare(o1.m_second, o2.m_second);
				}
			}
		);
		
		DataOutputStream out = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(filename)));
	
		int size = ranks.size();
	
		out.writeInt(size);
		for(Pair<Integer, Double> r : ranks)
		{
			out.writeInt(r.m_first);
			out.writeDouble(r.m_second);
		}
	
		out.close();
		
		System.out.println("PageRank: done");
	}

	/**
	 * Saves pagerank values to the specified file.
	 * 
	 * @param filename
	 *            file to be saved to
	 * @throws IOException
	 */
	public void saveToFile_HumanReadable(IdToURLMapping mapping, String filename) throws IOException
	{
		System.out.println("PageRank: Saving sorted humanfriendly ranks to file");
		
		ArrayList<Pair<Integer, Double> > ranks = new ArrayList<Pair<Integer, Double> >(Ranks.length);
		for(int i = 0; i < Ranks.length; ++i)
			ranks.add(new Pair<Integer, Double>(i, Ranks[i]));
		
		Collections.sort(ranks, new Comparator<Pair<Integer, Double>>() {
				@Override
				public int compare(Pair<Integer, Double> o1, Pair<Integer, Double> o2)
				{
					return o1.m_second.compareTo(o2.m_second);
				}
			}
		);
		
		BufferedWriter wr = new BufferedWriter(new FileWriter(filename));
		for(Pair<Integer, Double> r : ranks)
		{
			String url = mapping.get(r.m_first);
			if(url == null)
			{
				System.out.println("Unknown URL in graph file " + r.m_first);
				continue;
			}
			wr.write(url);
			wr.write("\t");
			wr.write(Double.toString(r.m_second));
			wr.write("\n");
		}
	
		wr.close();
		System.out.println("PageRank: done");
	}

	/**
	 * Saves pagerank values to the specified file in text (csv) format.
	 * 
	 * @param filename
	 *            file to be saved to
	 * @param domains	Mapping of domains-ids to domains
	 * @throws IOException
	 */
	public void saveToFile_HumanReadable(String[] domains, String filename) throws IOException
	{
		System.out.println("PageRank: Saving sorted humanfriendly ranks to file");
		
		ArrayList<Pair<Integer, Double> > ranks = new ArrayList<Pair<Integer, Double> >(Ranks.length);
		for(int i = 0; i < Ranks.length; ++i)
			ranks.add(new Pair<Integer, Double>(i, Ranks[i]));
		
		Collections.sort(ranks, new Comparator<Pair<Integer, Double>>() {
				@Override
				public int compare(Pair<Integer, Double> o1, Pair<Integer, Double> o2)
				{
					return o1.m_second.compareTo(o2.m_second);
				}
			}
		);
		
		BufferedWriter wr = new BufferedWriter(new FileWriter(filename));
		for(Pair<Integer, Double> r : ranks)
		{
			wr.write(domains[r.m_first]);
			wr.write("\t");
			wr.write(Double.toString(r.m_second));
			wr.write("\n");
		}
	
		wr.close();
		System.out.println("PageRank: done");
	}

	public double getRank(int node)
	{
		return Ranks[node];
	}

	/**
	 * Frees internal buffers in order to reduce memory footprint. Use this method
	 * after a pagerank-computation. The instance will not be able to do further operations
	 * except saving its data.
	 */
	public void strip()
	{
		to = null;
		PRold = null;
		PRnew = null;
	}

	/**
	 * Simple switch method for switching the pointers to the arrays holding the
	 * old pagerank values and the new ones in each iteration.
	 */
	protected void switchArrays()
	{
		double[] tmp;
		tmp = PRold;
		PRold = PRnew;
		PRnew = tmp;
	}
	
	
	public interface itfPr
	{ // whooohooo polymorp rox!!!
		/**
		 * Returns the outdegree of a node
		 * 
		 * @param id
		 *            - the node id
		 * @return the outdegree of the node
		 */
		public int getOutdegree(int id);
	
		/**
		 * Returns the indegree of a node
		 * 
		 * @param id
		 *            - the node id
		 * @return the outdegree of the node
		 */
		public int getIndegree(int id);
	
		/**
		 * Returns the inlist of a node
		 * 
		 * @param id
		 *            - the node id
		 * @return the node id array of the inlinks of the node
		 */
		public int[] inlist(int id);
	}
	
	protected class toServer implements itfPr
	{ // poly to server
		private int[] outdegree;
		private int[][] inlist;

		public int getOutdegree(int i)
		{
			return outdegree[i];
		}

		public int getIndegree(int i)
		{
			return inlist[i].length;
		}

		public int[] inlist(int i)
		{
			return inlist[i];
		}

		public toServer(int[] outdegree, int[][] inlist)
		{
			this.outdegree = outdegree;
		}
	}

	protected class toLocal implements itfPr
	{ // poly to local
		private KrakeWebGraph graph;

		public int getOutdegree(int i)
		{
			return graph.getOutdegree(i);
		}

		public int getIndegree(int i)
		{
			return graph.getIndegree(i);
		}

		public int[] inlist(int i)
		{
			return graph.getInlinks(i);
		}

		public toLocal(KrakeWebGraph graph)
		{
			this.graph = graph;
		}
	}
}
